#!/bin/bash

#usage: $o [experiment-id  title]+
# Creates a report about the given experiment id(s compared).

#settings file must contain the exports of the following variables:
# DETLOG  filename of the detailed log file. THe individual events are written into this log file
# TEMPDIR is a directory (without the trailing /) where to save temporary files (small files for communicating between processes of the script)


export DESCLEN=30
export NUMLEN=16

# read settings
if [ -z "$FIMIDIR" ] ; then
    if . ./settings ; then
        x=1 # do nothing
    else
        echo Error: either create ./settings or set FIMIDIR variable to the directory of it >&2
        exit 1
    fi
else 
    if . $FIMIDIR/settings ; then
        x=1
    else
        echo Error: $FIMIDIR/settings not found. Set FIMIDIR variable correctly. >&2
        exit 1
    fi
fi

export idx=0
IDS=()
DESCS=()
while [ ! -z "$1" ] ; do
    IDS[$idx]=$1
    shift
    DESCS[$idx]=$1    
    shift
    idx=$(($idx + 1))
done

export IDS DESCS

#get an experiment result
# arg 1 is the experiment id
# arg 2 is the result key
# arg 3 [optional] is the pass id. If not specified, then the first match is taken.
# the resulting number is printed on stdout. An error is printed on stderr if no match is found
function getresult() {
    if grep -q "^$1:$2:$3" $DETLOG ; then
        grep "^$1:$2:$3" $DETLOG | cut -d : -f 4 | head -n 1 
    else
        echo Error: experiment id $1 run $3 result $2 not found. >&2
        echo 0
    fi
}

#get an experiment result
# arg 1 is the experiment id
# arg 2 is the result key
# all matching results are printed on stdout one at a line (only the numbers)
function getallresult() {
    grep "^$1:$2:" $DETLOG | cut -d : -f 4
}


function getcurrresult() {
    getresult ${IDS[$i]} "$@"
}

function getcurrallresult() {
    getallresult ${IDS[$i]} "$@"
}

# arg 1 : a number
# stdout: the same number in decimal format with , as the thousand separator
function formatnumber() {
    NUM=$1
#    REP=`printf "%03d" $(( $NUM % 1000 ))`
    REP=""
    while [ $NUM -ge 1000 ] ; do
        REP=`printf ",%03d%s" $(( $NUM % 1000 )) "$REP" `
        NUM=$(( $NUM / 1000 ))
    done
    printf "%d%s" $(( $NUM % 1000 )) "$REP"
}

# arg 1: the description to be printed at the beginnning of the line
# arg 2: the command that returns the value to print. SHould use index variable i to identify the data source index.
function displine() {
    DESC=$1
    DESC=${DESC:0:$DESCLEN}
    rem=$(( $DESCLEN - ${#DESC} ))
    if (($latex)) ; then
        echo -n "$DESC"
    else
        printf "%s%${rem}s " "$DESC" ""
    fi
    for((i=0; $i<$idx; i++)) do
       export i
       #echo $2 "$2"
       NUM=`$2`
       #echo num=$NUM
#       NUM=`getresult ${IDS[$i]} $2 $3`
       NF=`formatnumber $NUM`
       if (($latex)) ; then
           printf "&%s" $NF
       else
           printf "%${NUMLEN}s " $NF
       fi
    done
    if (($latex)) ; then
        echo '\\'
    else
        echo
    fi
}


function getavgtsc() {
    getcurrallresult tsc | awk 'BEGIN {sum=0;count=0;} {sum+=$1; count++;} END {printf("%d\n",sum/count);}'
}

function getdifftsc() {
    getcurrallresult tsc | awk -v avg=`getavgtsc` ' function abs(i) {return i<0?-i:i;}  BEGIN {sum=0;count=0;} {sum+=abs($1-avg);count++;} END{printf("%d\n",sum/count);}'
}

# return arg 1 - arg 2
# the arguments may contain one space to fill the optional run argument of getcurrresult
function getcurrdiff() {
    echo $(( `getcurrresult $1` - `getcurrresult $2` ))
}


function getdiffpercent() {
    echo $(( (`getcurrresult $1` - `getcurrresult $2`) / (`getcurrresult $3` / 100000 ) ))
}

function getpercent() {
    echo $(( (`getcurrresult $1` ) / (`getcurrresult $2` / 100000 ) ))
}


function header() {
    if (($latex)) ; then
        echo '\hline\hline' 
        echo '\multicolumn{3}{l}{\textbf{'"$1"'}}\\'
    else
        echo
        echo '---------------------------------'
        echo $1
        echo
    fi
}

function blankline() {
    if (($latex)) ; then
        echo '\hline'
    else
        echo
    fi
}

if (($latex)) ; then

    echo -n '\begin{tabular}{l'
    for ((i=0;i<idx;++i)) ; do
        echo -n 'r'
    done
    echo '}'

    echo -n "% Experiment IDs: "
    printf "%s " ${IDS[*]}
    echo
    
    echo '\multicolumn{2}{l}{FULL EXECUTION PROFILE}\\'
    printf "&%s" ${DESCS[*]}
    echo '\\'
    
else
    echo 'FULL EXECUTION PROFILE'
    echo
    echo
    echo -n "Experiment IDs: "
    printf "%s " ${IDS[*]}
    echo
    printf "%${DESCLEN}s" ""
    printf "%${NUMLEN}s" ${DESCS[*]}

fi

header 'Time'
displine "tsc avg" "getavgtsc"
displine "+-" "getdifftsc"

header 'Execution'
displine "Instructions retired" "getcurrresult instr_all"
displine "u-op retired" "getcurrresult uop_all"
#displine "Instructions bogus" "getcurrresult instr_bogus"
#displine "uops nbogus" "getcurrresult uops_nbogus"
#displine "uops bogus" "getcurrresult uops_bogus"
#blankline
#displine 'uop from TC build' "getcurrresult uop_from_TC_build"
#displine 'uop from TC deliver' "getcurrresult uop_from_TC_deliver"
#displine 'uop from ROM' "getcurrresult uop_from_ROM"

header 'Load-Store events'
displine "LS2 buffer full" "getcurrresult LS2_buffer_full"
displine "WC buffer flush" "getcurrdiff WC_buffer_flush_plus_streaming_store streaming_store"
displine "streaming store" "getcurrresult streaming_store"
blankline
displine "prefetch instructions" "getcurrresult prefetch_instr"
displine "HW prefetch L2cache hits" "getcurrresult hw_prefetch_cachehit"
displine "HW prefetch" "getcurrresult hw_prefetch_all"
blankline
displine "64bit writes to memory" "getcurrresult FSB_write_8byte"

header 'Cache efficiency'
displine "L1 data access" "getcurrresult L1D_access"
displine "L1D miss L2 hit" "getcurrresult L1D_miss_L2_hit"
displine "L1D miss L2 miss" "getcurrresult L1D_miss_L2_miss"
displine "L1D misaligned acc" "getcurrresult L1D_misalign_access"
blankline
displine "L2 cache misses (all)" "getcurrresult L2_miss_all"
displine "L2 TLB walk misses" "getcurrresult L2_miss_TLB_walk"


header 'TLB'
displine "L1DTLB miss L2DTLB hit" "getcurrresult L1DTBL_miss_L2DTBL_hit"
displine "L1DTLB miss L2DTLB miss" "getcurrresult L1DTBL_miss_L2DTBL_miss"

header 'Miscellaneous'
displine "Far control transfers" "getcurrresult far_control_transfer"
displine "mmx operations" "getcurrresult mmx_oper"
displine "ticks not halted (user time)" "getcurrresult uop_histo_all_3"

header 'Branches'
displine "All branches" "getcurrresult branch_all"
displine "Mispredicted branches" "getcurrresult branch_mispred"

header 'Stalls'
displine "Decoder input empty" "getcurrresult stall_decoder_empty"
displine "Decoder output stall" "getcurrresult stall_dispatch"
displine "Branch abort wait" "getcurrresult stall_branch_abort"
blankline
displine "FPU full (ovr=may overlap)" "getcurrresult stall_fpu_full"
displine "Reorder buffer full (ovr)" "getcurrresult stall_reorder_buffer_full"
displine "Reservation station full (ovr)" "getcurrresult stall_resrvstat_full"
displine "Load/Store unit full" "getcurrresult stall_LS_full"

header 'Memory controller (DRAM events)'
displine "Page hit" "getcurrresult dram_page_hit"
displine "Page miss" "getcurrresult dram_page_miss"
displine "Page conflict" "getcurrresult dram_page_conflict"
displine "Page table overflows" "getcurrresult dram_page_table_overflows"
displine "Turnarounds" "getcurrresult dram_turnaround_all"

header 'Execution histogram'
displine "TickOf uop stall" "getcurrresult uop_histo_all_0"
displine "TickOf uop 1" "getcurrdiff uop_histo_all_1 uop_histo_all_0"
displine "TickOf uop 2" "getcurrdiff uop_histo_all_2 uop_histo_all_1"
displine "TickOf uop 3" "getcurrdiff uop_histo_all_3 uop_histo_all_2"
blankline
displine "Percent uop stall" "getpercent uop_histo_all_0 uop_histo_all_3"
displine "Percent uop 1" "getdiffpercent uop_histo_all_1 uop_histo_all_0 uop_histo_all_3"
displine "Percent uop 2" "getdiffpercent uop_histo_all_2 uop_histo_all_1 uop_histo_all_3"
displine "Percent uop 3" "getdiffpercent uop_histo_all_3 uop_histo_all_2 uop_histo_all_3"


if (($latex)); then
    echo '\end{tabular}'
fi

exit

