#!/bin/bash

#usage: $o [experiment-id  title]+
# Creates a report about the given experiment id(s compared).

#settings file must contain the exports of the following variables:
# DETLOG  filename of the detailed log file. THe individual events are written into this log file
# TEMPDIR is a directory (without the trailing /) where to save temporary files (small files for communicating between processes of the script)


export DESCLEN=30
export NUMLEN=16

# read settings
if [ -z "$FIMIDIR" ] ; then
    if . ./settings ; then
        x=1 # do nothing
    else
        echo Error: either create ./settings or set FIMIDIR variable to the directory of it >&2
        exit 1
    fi
else 
    if . $FIMIDIR/settings ; then
        x=1
    else
        echo Error: $FIMIDIR/settings not found. Set FIMIDIR variable correctly. >&2
        exit 1
    fi
fi

export idx=0
IDS=()
DESCS=()
while [ ! -z "$1" ] ; do
    IDS[$idx]=$1
    shift
    DESCS[$idx]=$1    
    shift
    idx=$(($idx + 1))
done

export IDS DESCS

#get an experiment result
# arg 1 is the experiment id
# arg 2 is the result key
# arg 3 [optional] is the pass id. If not specified, then the first match is taken.
# the resulting number is printed on stdout. An error is printed on stderr if no match is found
function getresult() {
    TMPF=$TEMPDIR/resf.$$
#    TMPF=$TEMPDIR/resf.$$.`sed s/:/_/g <<< $2`.$3
    rm -f $TMPF
    if grep "^$1:$2:$3" $DETLOG > $TMPF ; then
        cut -d : -f 4 $TMPF | head -n 1 
    else
        echo Error: experiment id $1 run $3 result $2 not found. >&2
        echo 0
    fi
    rm -f $TMPF
}


function getcurrresult() {
    getresult ${IDS[$i]} "$@"
}

# arg 1 : a number
# stdout: the same number in decimal format with , as the thousand separator
function formatnumber() {
    NUM=$1
#    REP=`printf "%03d" $(( $NUM % 1000 ))`
    REP=""
    while [ $NUM -ge 1000 ] ; do
        REP=`printf ",%03d%s" $(( $NUM % 1000 )) "$REP" `
        NUM=$(( $NUM / 1000 ))
    done
    printf "%d%s" $(( $NUM % 1000 )) "$REP"
}

# arg 1: the description to be printed at the beginnning of the line
# arg 2: the command that returns the value to print. SHould use index variable i to identify the data source index.
function displine() {
    DESC=$1
    DESC=${DESC:0:$DESCLEN}
    rem=$(( $DESCLEN - ${#DESC} ))
    if (($latex)) ; then
        echo -n "$DESC"
    else
        printf "%s%${rem}s " "$DESC" ""
    fi
    for((i=0; $i<$idx; i++)) do
       export i
       #echo $2 "$2"
       NUM=`$2`
       #echo num=$NUM
#       NUM=`getresult ${IDS[$i]} $2 $3`
       NF=`formatnumber $NUM`
       if (($latex)) ; then
           printf "&%s" $NF
       else
           printf "%${NUMLEN}s " $NF
       fi
    done
    if (($latex)) ; then
        echo '\\'
    else
        echo
    fi
}


function getavgtsc() {
    echo $(( (`getcurrresult tsc 1` + `getcurrresult tsc 2` + `getcurrresult tsc 3`) / 3 ))
}

function getdifftsc() {
    R1=`getcurrresult tsc 1`
    R2=`getcurrresult tsc 2`
    R3=`getcurrresult tsc 3`
    AVG=$(( ( R1 + R2 + R3 ) / 3 ))
    A1=$(( R1 - AVG ))
    A2=$(( R2 - AVG ))
    A3=$(( R3 - AVG ))   
    
    echo $(( (A1 < 0 ? -A1 : A1 ) + (A2 < 0 ? -A2 : A2 ) + (A3 < 0 ? -A3 : A3 ) )) 
}

# return arg 1 - arg 2
# the arguments may contain one space to fill the optional run argument of getcurrresult
function getcurrdiff() {
    echo $(( `getcurrresult $1` - `getcurrresult $2` ))
}


function getdiffpercent() {
    echo $(( (`getcurrresult $1` - `getcurrresult $2`) / (`getcurrresult $3` / 100000 ) ))
}

function getpercent() {
    echo $(( (`getcurrresult $1` ) / (`getcurrresult $2` / 100000 ) ))
}


function header() {
    if (($latex)) ; then
        echo '\hline\hline' 
        echo '\multicolumn{3}{l}{\textbf{'"$1"'}}\\'
    else
        echo
        echo '---------------------------------'
        echo $1
        echo
    fi
}

function blankline() {
    if (($latex)) ; then
        echo '\hline'
    else
        echo
    fi
}

if (($latex)) ; then

    echo -n '\begin{tabular}{l'
    for ((i=0;i<idx;++i)) ; do
        echo -n 'r'
    done
    echo '}'

    echo -n "% Experiment IDs: "
    printf "%s " ${IDS[*]}
    echo
    
    echo '\multicolumn{2}{l}{FULL EXECUTION PROFILE}\\'
    printf "&%s" ${DESCS[*]}
    echo '\\'
    
else
    echo 'FULL EXECUTION PROFILE'
    echo
    echo
    echo -n "Experiment IDs: "
    printf "%s " ${IDS[*]}
    echo
    printf "%${DESCLEN}s" ""
    printf "%${NUMLEN}s" ${DESCS[*]}

fi

header 'Time'
displine "tsc avg" "getavgtsc"
displine "+-" "getdifftsc"

header 'Execution'
displine "Instructions nbogus" "getcurrresult instr_nbogus"
displine "Instructions bogus" "getcurrresult instr_bogus"
displine "uops nbogus" "getcurrresult uops_nbogus"
displine "uops bogus" "getcurrresult uops_bogus"
blankline
displine 'uop from TC build' "getcurrresult uop_from_TC_build"
displine 'uop from TC deliver' "getcurrresult uop_from_TC_deliver"
displine 'uop from ROM' "getcurrresult uop_from_ROM"

header 'FSB/Memory events'
displine "Count of writes" "getcurrresult FSB_write"
displine "Count of reads (incl. prefetch)" "getcurrresult FSB_read_plus_prefetch"
displine "Count of r+w+prefetch" "getcurrresult fsb_r_w_prefetch"
displine "Ticks of r+w+prefetch pending" "getcurrresult fsb_tickof_r_w_prefetch_pending"
displine "Count of r+w" "getcurrresult fsb_r_w"
displine "Ticks of r+w pending" "getcurrresult fsb_tickof_r_w_pending"
blankline
displine "store operations (nbogus)" "getcurrresult store_nbogus"
displine "load operations (nbogus)" "getcurrresult load_nbogus"
displine "128bit mmx operations" "getcurrresult 128bit_mmx_oper"

header 'Branches'
displine "Function calls" "getcurrresult total_function_calls"
displine "Indirect branches" "getcurrresult branch_indirect"
blankline
displine "Total conditional branches" "getcurrresult branch_all_conditional"
displine "Mispred cond branches" "getcurrresult branch_mispred_conditional"
displine "Mispred noncond branches (?)" "getcurrresult mispred_noncond_branch"

header 'Stall causes'
displine "MemoryOrderBuffer load (pcs)" "getcurrresult MOB_load_replay"
displine "Lack of store buffer" "getcurrresult resource_stall_store_buffer"
displine "Memory cancel" "getcurrresult memory_cancel_all"
displine "Split memory access" "getcurrresult memory_complete_split_access"
displine "WC buffer evictions" "getcurrresult wc_buffer_evictions"
blankline
displine "L1 read miss" "getcurrresult L1_read_miss"
displine "L2 read miss" "getcurrresult L2_read_miss"

header 'Execution histogram'
displine "TickOf uop stall" "getcurrresult uop_histo_all_0"
displine "TickOf uop 1" "getcurrdiff uop_histo_all_1 uop_histo_all_0"
displine "TickOf uop 2" "getcurrdiff uop_histo_all_2 uop_histo_all_1"
displine "TickOf uop 3" "getcurrdiff tsc:3 uop_histo_all_2"
blankline
displine "TickOf nbogus uop stall" "getcurrresult uop_histo_nbogus_0"
displine "TickOf nbogus uop 1" "getcurrdiff uop_histo_nbogus_1 uop_histo_nbogus_0"
displine "TickOf nbogus uop 2" "getcurrdiff uop_histo_nbogus_2 uop_histo_nbogus_1"
displine "TickOf nbogus uop 3" "getcurrdiff tsc:3 uop_histo_nbogus_2"

blankline
displine "Percent uop stall" "getpercent uop_histo_all_0 tsc:3"
displine "Percent uop 1" "getdiffpercent uop_histo_all_1 uop_histo_all_0 tsc:3"
displine "Percent uop 2" "getdiffpercent uop_histo_all_2 uop_histo_all_1 tsc:3"
displine "Percent uop 3" "getdiffpercent tsc:3 uop_histo_all_2 tsc:3"
blankline
displine "Percent nbogus uop stall" "getpercent uop_histo_nbogus_0 tsc:3"
displine "Percent nbogus uop 1" "getdiffpercent uop_histo_nbogus_1 uop_histo_nbogus_0 tsc:3"
displine "Percent nbogus uop 2" "getdiffpercent uop_histo_nbogus_2 uop_histo_nbogus_1 tsc:3"
displine "Percent nbogus uop 3" "getdiffpercent tsc:3 uop_histo_nbogus_2 tsc:3"

if (($latex)); then
    echo '\end{tabular}'
fi

exit

CTRS=""
DESC=""

CTRS="$CTRS -e 0x0103d000/0x06508204@0x80000001"
DESC="$DESC FSB_write"
CTRS="$CTRS -e 0x0103d000/0x07504204@0x80000002"
DESC="$DESC FSB_read_plus_prefetch"
CTRS="$CTRS -e 0x00031000/0x12000204@0x80000004"
DESC="$DESC uop_from_TC_build"
CTRS="$CTRS -e 0x00031000/0x12000404@0x80000006"
DESC="$DESC uop_from_TC_deliver"
CTRS="$CTRS -e 0x00035000/0x0a003804@0x80000005"
DESC="$DESC mispred_noncond_branch"
CTRS="$CTRS -e 0x00035000/0x08000804@0x80000007"
DESC="$DESC total_function_calls"
CTRS="$CTRS -e 0x00033000/0x02004004@0x8000000c"
DESC="$DESC resource_stall_store_buffer"
CTRS="$CTRS -e 0x0003b000/0x10000204@0x80000010"
DESC="$DESC store_nbogus"
CTRS="$CTRS -e 0x00035000/0x04000804@0x80000011"
DESC="$DESC tag_store1"
CTRS="$CTRS -e 0x0003B000/0x12000604@0x8000000E --p4pe=0x01000001 --p4pmv=0x1"
DESC="$DESC L1_read_miss"
CTRS="$CTRS -e 0x00039000/0x02000204@0x8000000d"
DESC="$DESC uops_nbogus"
CTRS="$CTRS -e 0x00039000/0x04000604@0x8000000f"
DESC="$DESC instr_nbogus"


runprog "$CTRS" 1 "$DESC"

CTRS=""
DESC=""

CTRS="$CTRS -e 0x0003b000/0x04001804@0x80000008" 
DESC="$DESC memory_cancel_all"
CTRS="$CTRS -e 0x00035000/0x10000604@0x80000009"
DESC="$DESC memory_complete_split_access"
CTRS="$CTRS -e 0x0003b000/0x0a000604@0x8000000a"
DESC="$DESC wc_buffer_evictions"
CTRS="$CTRS -e 0x00033000/0x35000004@0x8000000b"
DESC="$DESC 128bit_mmx_oper"
CTRS="$CTRS -e 0x00031000/0x12000804@0x80000005"
DESC="$DESC uop_from_ROM"
CTRS="$CTRS -e 0x0103d000/0x0750c204@0x80000000"
DESC="$DESC fsb_r_w_prefetch"
CTRS="$CTRS -e 0x0007d000/0x3550c204@0x80000002"
DESC="$DESC fsb_tickof_r_w_prefetch_pending"
CTRS="$CTRS -e 0x00035000/0x06007404@0x80000001"
DESC="$DESC MOB_load_replay"
CTRS="$CTRS -e 0x00035000/0x0a000404@0x80000005"
DESC="$DESC branch_mispred_conditional"
CTRS="$CTRS -e 0x00035000/0x08000404@0x80000006"
DESC="$DESC branch_all_conditional"
CTRS="$CTRS -e 0x0003B000/0x10000204@0x8000000C"
DESC="$DESC load_nbogus"
CTRS="$CTRS -e 0x00035000/0x04000404@0x8000000D"
DESC="$DESC tag_load2"
CTRS="$CTRS -e 0x0003B000/0x12000604@0x8000000E --p4pe=0x01000002 --p4pmv=0x1"
DESC="$DESC L2_read_miss"
CTRS="$CTRS -e 0x00039000/0x02000404@0x8000000F"
DESC="$DESC uops_bogus"
CTRS="$CTRS -e 0x00039000/0x04001804@0x80000010"
DESC="$DESC instr_bogus"

runprog "$CTRS" 2 "$DESC"

CTRS=""
DESC=""


#for run 3
CTRS="$CTRS -e 0x00035000/0x08002004@0x80000007"
DESC="$DESC branch_indirect"

CTRS="$CTRS -e 0x000f9000/0x02000604@0x8000000c"
DESC="$DESC uop_histo_all_0"
CTRS="$CTRS -e 0x001f9000/0x02000604@0x8000000d"
DESC="$DESC uop_histo_all_1"
CTRS="$CTRS -e 0x002f9000/0x02000604@0x80000010"
DESC="$DESC uop_histo_all_2"
CTRS="$CTRS -e 0x000f9000/0x02000204@0x8000000e"
DESC="$DESC uop_histo_nbogus_0"
CTRS="$CTRS -e 0x001f9000/0x02000204@0x8000000d"
DESC="$DESC uop_histo_nbogus_1"
CTRS="$CTRS -e 0x002f9000/0x02000204@0x80000011"
DESC="$DESC uop_histo_nbogus_2"


runprog "$CTRS" 3 "$DESC"