/*!
  \file   Apriori.hpp
  \brief  This class implements the APRIORI algorithm.

  \author Ferenc BODON
  \date   2005-04-02
*/



#ifndef Apriori_HPP
#define Apriori_HPP

#include "common.hpp"
#include "common/log.h"
#include "common/allocators.hpp"
#include <vector>

/** 
\brief This class implements the APRIORI algorithm.

<p>
APRIORI is a levelwise algorithm.
It scans the transaction database several times.
After the first scan the frequent 1-itemsets are found, and in general 
after the <em>k<sup>th</sup></em> scan the frequent <em>k</em>-itemsets 
are extracted. The method does not determine the support of every possible
itemset. In an attempt to narrow the domain to be searched, before every pass 
it generates <em>candidate</em> itemsets. An itemset becomes a candidate 
if every subset of it is frequent. Obviously every frequent itemset 
needs to be candidate too, hence only the support of candidates is calculated.
Frequent <em>k</em>-itemsets generate the candidate <em>k+1</em>-itemsets 
after the <em>k<sup>th</sup></em> scan.
</p>

<p>
After all the candidate <em>(k+1)</em>-itemsets have been generated, a new 
scan of the transactions is effected and the precise support of the 
candidates is determined. The candidates with low support are thrown away. The 
algorithm ends when no candidates can be generated.
</p>

<p>
The intuition behind candidate generation is based on the following simple 
fact:<br><div align="center"><em>Every subset of a frequent itemset 
is frequent.</em></div><br> This is immediate, because if a transaction 
<em>t</em> supports an itemset <em>X</em>, then <em>t</em> supports 
every subset \f$Y\subseteq X\f$.
</p>

<p>
Using the fact indirectly, we infer, that if an itemset has a subset that is 
infrequent, then it cannot be frequent. So in the algorithm APRIORI only those 
itemsets will be candidates whose every subset is frequent. The frequent 
<em>k</em>-itemsets are available when we attempt to generate candidate 
<em>(k+1)</em>-itemsets. The algorithm seeks candidate <em>k+1</em>-itemsets 
among the sets which are unions of two frequent <em>k</em>-itemsets. After 
forming the union we need to verify that all of its subsets are frequent, 
otherwise it should not be a candidate. To this end, it is clearly enough to 
check if all the <em>k</em>-subsets of <em>X</em> are frequent.
</p>

<p>
Next the supports of the candidates are calculated. This is done by reading 
transactions one by one. For each transaction <em>t</em> the algorithm decides 
which candidates are supported by <em>t</em>. To solve this task efficiently 
it is useful to store the candidates in a special datastructure like trie or 
hash-tree. 
</p>
The template parameters are the following:
   \param C type of the coder
   \param D type of the decoder
   \param MAIN_DATA_STRUCTURE the main data structure. It is basically a trie (prefix-trie)
   \param SEC_DATA_STRUCTURE the secondary data structure
   \param FII the type of a frequent item inserter (it inserts frequent items into the data structure)
   \param FPI the type of a frequent item inserter (it inserts frequent pairs into the data structure)
   \param CG the type of candidate generator
   \param IR the type of the infrequent remover
   \param SC the type of the support counter
*/
template <class C, class D, class MAIN_DATA_STRUCTURE, 
	  class SEC_DATA_STRUCTURE, class FII, class FPI, 
	  class CG, class IR, class SC>
class Apriori
{
   private:
      /// the data strucure that stores candidate.
      MAIN_DATA_STRUCTURE& data_structure;
      /// The coder
      C& coder;
      /// The decoder
      D& decoder;
      //
      FII& fii;

      unsigned int candidate_size;

   public:

      Apriori(MAIN_DATA_STRUCTURE& data_structure, C& coder, D& decoder, FII& fii ) : 
	 data_structure(data_structure), coder(coder), decoder(decoder), fii(fii){}

      /// This method finds the frequent itemsets
      void findFrequentItemsets( 
	 const counter_t emptyset_support,
	 std::vector<counter_t>& freq_counters,
	 std::vector< std::pair< counter_t, std::pair<item_t, item_t> > >& 
	 freq_pairs_with_counters,
	 const counter_t min_supp,
	 unsigned int maxsize = largest_itemsetsize);

};

/**
   \param emptyset_support the support of the emptyset (i.e. number of
   transaction) 
   \param freq_counters a vector that stores the support
   of the frequent items. freq_counters[i] belongs to item \e i.   
   \param freq_pairs_with_counters a vector that stores the support 
   of the frequent pairs. The elements of the vector are pairs, where 
   the first elemen stores the support value and the second element stores
   the pair itself.
   \param min_supp the support threshold
   \param maxsize the maximal size of the frequent itemset. If this parameter
   is set to 0 then there is no limit for the maximal size.
 */
template <class C, class D, class MAIN_DATA_STRUCTURE, 
	  class SEC_DATA_STRUCTURE, class FII, class FPI, 
	  class CG, class IR, class SC>
void Apriori<C, D, MAIN_DATA_STRUCTURE, SEC_DATA_STRUCTURE, 
	     FII, FPI, CG, IR, SC>::
findFrequentItemsets( 
   const counter_t emptyset_support,
   std::vector<counter_t>& freq_counters,
   std::vector< std::pair< counter_t, std::pair<item_t, item_t> > >& 
   freq_pairs_with_counters,
   const counter_t min_supp,
   unsigned int maxsize )
{
   if(maxsize > 2)
   {
      SEC_DATA_STRUCTURE  s_alloc;

      FPI fpi(data_structure, decoder, s_alloc);
      log_status(0,"Setting the support of the emptyset in the trie.");
      fii.setEmptysetSupport(emptyset_support);
      log_status(0,"Inserting the support of the frequent items.");
      fii.insertFrequentItems(freq_counters);
      freq_counters.clear();
      std::vector<counter_t>().swap(freq_counters);
      log_status(0,"Inserting the frequent pairs and they supports.");
      fpi.insertFrequentPairs(freq_pairs_with_counters);
      freq_pairs_with_counters.clear();  
      std::vector< std::pair< counter_t, std::pair<item_t, item_t> > >().
	 swap(freq_pairs_with_counters);

      CG candidate_generator(data_structure, decoder, s_alloc);
      SC support_counter(data_structure, coder);
      IR infrequent_remover(data_structure, decoder, s_alloc);

      log_status(0,"Setting the size of the candidate to 3");
      candidate_size = 3;

      log_status(0,"Generating candidates of size 3");
      candidate_generator.generateCandidate(candidate_size);
   
      while( candidate_generator.isThereAnyCandidate()  )
      {
	 log_status(0,"Determining the supports");
	 support_counter.countSupport(candidate_size);
	 log_status(0,"Deleting infrequent candidates");
	 infrequent_remover.deleteInfrequent(min_supp, candidate_size);
	 ++candidate_size;
	 log_status(0,"Setting the candidatesize counter to %d", 
		    candidate_size);
	 if(maxsize < candidate_size)
	 {
	    log_status(0,"Maximal size threshold is reached.");
	    break;
	 }
	 log_status(0,"Generating new candidates");
	 candidate_generator.generateCandidate(candidate_size);
      }
      log_status(0,"No new candidates were generated.");
      candidate_generator.afterWorkCandGen();
      infrequent_remover.afterWorkDel();
   }
   else
   {
      for(std::vector<counter_t>::size_type index = 0;
	  index != freq_counters.size(); ++index)
      {
	 decoder.pushItemWithWrite( index, freq_counters[index] );
	 decoder.popItem();	 
      }
      for(std::vector< std::pair< counter_t, std::pair<item_t, item_t> > >::iterator it 
	     = freq_pairs_with_counters.begin(); it != freq_pairs_with_counters.end(); ++it)
      {
	 // it could be solved more efficient... but I don't care ;)
	 decoder.pushItem((*it).second.first);
	 decoder.pushItemWithWrite((*it).second.second, (*it).first);
	 decoder.popItem();
	 decoder.popItem();
      }      
   }
}

#endif
