/*!
  \file   AprioriSelector.hpp
  \brief  This class processes the command line
  option and creates the proper objects that are needed by Apriori.

  \author Ferenc BODON
  \date   2005-05-02
*/

#ifndef AprioriSelector_HPP
#define AprioriSelector_HPP

#include "common/allocators.hpp"


#include "apriori/bodon/dynamic_trie/trie_manipulators/FrequentItemInserter.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/FrequentPairInserter.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/FrequentPairInserterNoprune.hpp"

#include "apriori/bodon/dynamic_trie/trie_manipulators/IntersectProPruner.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/CandidateGeneratorPrune.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/CandGenInfreqRemoveNopruneMerge.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/InfreqRemover.hpp"

#include "apriori/Apriori.hpp"

template <class TRIE_OEL, class TRIE_OI, class LEAF_WC, class S_C, class F2F, 
	  class SUPP_C, class T_R, class DF_D, NEELevel NEE = NEE_Full> 
class AprioriSelector
{
   public:
      AprioriSelector( 
	 counter_t min_supp, char* algorithm, char* input_file,
	 counter_t nr_of_transactions, 
	 std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
	 T_R& tr_reader, DF_D& df_decoder, unsigned int maxsize = largest_itemsetsize);
};

template <class TRIE_OEL, class TRIE_OI, class LEAF_WC, class S_C, class F2F, 
	  class SUPP_C, class T_R, class DF_D, NEELevel NEE>
AprioriSelector<TRIE_OEL, TRIE_OI, LEAF_WC, S_C, F2F, SUPP_C, T_R, DF_D, NEE>::AprioriSelector( 
   counter_t min_supp, char* algorithm, char* input_file,
   counter_t nr_of_transactions, 
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters, 
   T_R& tr_reader, DF_D& df_decoder, unsigned int maxsize )
{
   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = ASC;
//   par_c.codemode = DESC;
   log_status(0,"Doing sorted codec.");
   S_C sorted_coder(&par_c);

   std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
      freq_pairs_with_counters;
   F2F fr_2_filter(&sorted_coder );
   log_status(0,"Finding frequent pairs.")
      fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);

   TRIE_OEL main_trie;
   typedef Bodon::dynamic_trie::FrequentItemInserter<DF_D, TRIE_OEL, NEE> FII;
   FII fii(main_trie, df_decoder);
   typedef bracz::singleualloc<LEAF_WC, 64 * 1024> LEAF_ALLOCATOR;
   char *kind=algorithm+6;


   if( strstr(kind,"-noprune") )
   {
      log_status(0,"Complete pruning is disabled.");
      typedef Bodon::dynamic_trie::FrequentPairInserterNoprune<
	 DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::dynamic_trie::CandGenInfreqRemoveNopruneMerge<
	 DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE > CGIR;

      typedef Apriori<S_C, DF_D, TRIE_OEL, LEAF_ALLOCATOR, FII, 
	 FPI, CGIR, CGIR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.")
	 apriori.findFrequentItemsets( 
	    nr_of_transactions, *par_c.freq_counters,
	    freq_pairs_with_counters, min_supp, maxsize );
   }
   else
   {
      log_status(0,"Intersection-based pruning is selected.");
      typedef Bodon::dynamic_trie::FrequentPairInserter<DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, 
	 LEAF_ALLOCATOR, NEE> FPI;
      typedef Bodon::dynamic_trie::IntersectProPruner<
	 DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE> PRUNER;
      typedef Bodon::dynamic_trie::CandidateGeneratorPrune<PRUNER, DF_D, 
	 TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE> CG;
      typedef Bodon::dynamic_trie::InfreqRemover<
	 DF_D, TRIE_OEL, TRIE_OI, LEAF_WC, LEAF_ALLOCATOR, NEE> IR;
      typedef Apriori<S_C, DF_D, TRIE_OEL, LEAF_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
      A apriori(main_trie, sorted_coder, df_decoder, fii);
      log_status(0,"Finding frequent itemsets.");
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp, maxsize );
   }
}

#endif
