/*!
  \file   allocators.hpp
  \brief  Several different optimized memory allocation templates

  \author RACZ Balazs
  \date   2005-04-02
*/

#ifndef ALLOCATORS_HPP_050402
#define ALLOCATORS_HPP_050402

#include <list>
#include <vector>
#include <stack>

#include <malloc.h>
#include <string.h>
#include <stdlib.h>

#ifdef __SSE__
#include <xmmintrin.h>
#endif

#include "common/log.h"

#include <inttypes.h>




/// This is just a wrapper around the new and delete commands.
template<class T> class NewWrapperAlloc
{
  public:
   NewWrapperAlloc(){}
   T* DINLINE allocate(){
      return new T();
   }
   void free(T* data){
      delete data;
   }   
};

namespace bracz {
  class alignalloc {
    //must be a power of 2
    static const size_t alignment=16;
#ifdef __SSE__
    //    typedef int _v2q_t __attribute__((__mode__(V2DI)));
    typedef __m128i _v2q_t;

    static inline void DINLINE mymemzeroa(void *p,size_t len) {
      register int l=((len)+15)>>4;
      register _v2q_t a = {0,0};
      register _v2q_t *b=(_v2q_t*)p;
      while (l) {
        *b=a;
        ++b;
        --l;
      }
      //  memset(p,0,len<<2);
    }

    static inline void DINLINE mymemzerola(void *p,size_t len) {
      register int l=((len))>>4;
      register _v2q_t a = {0,0};
      register _v2q_t *b=(_v2q_t*)p;
      while (l) {
        *b=a;
        ++b;
        --l;
      }
      //zero remaining
      if (len & 15) {
        bzero(b,len & 15);
      }
    }

#else 
    #warning Enable SSE instructions to gain speed.

#define mymemzeroa bzero
#define mymemzerola bzero
    
#endif
  public:

    //! allocate an aligned array (without zeroing)
    template<class T> static void DINLINE alloc(T*& array, size_t size) {
      size_t tots=size*sizeof(T);
      //round up the length to a multiple of alignment
      tots+=2*(alignment-1)+sizeof(void*);
      void * base=malloc(tots);
      if(!base) {
        log_err(0,"Memory allocation failed.");
        abort();
      }
      //skip the size of a void* and align
      uintptr_t nb = reinterpret_cast<uintptr_t>(base);
      nb+=sizeof(void*)+alignment-1;
      nb&=~(alignment-1);
      array=reinterpret_cast<T*>(nb);
      //make place for the saved pointer
      void** saveptr=reinterpret_cast<void**>(nb);
      saveptr--;
      //save the original pointer
      *saveptr=base;
    }

    //! allocate and zero an aligned array
    template<class T> static void DINLINE allocz(T*& array, size_t size) {
      alloc(array,size);
      mymemzeroa(array,size*sizeof(T));
      //round up the length to a multiple of alignment
      //zero the array
      // **** MMX/SSE memory zero routine!
      //      memset(array,0,tots);
      //      bzero(array,tots);
    }
    

    //! zero an aligned array. The routine may zero memory after array+size up to the alignment.
    template<class T> static inline void DINLINE zero(T* array, size_t size) {
      // **** MMX/SSE memory zero routine!
      //      memset(array,0,tots);
      //      bzero(array,tots);        
      mymemzeroa(array,size*sizeof(T));
    }

    //! free an aligned array
    template<class T> static void NOINLINE free(T* array) {
      //look up the saved pointer
      void** ob = reinterpret_cast<void**>(array);
      ob--;
      //free
      ::free(*ob);
    }

    //! free an aligned array if it is not NULL
    template<class T> static void DINLINE freeifnonnull(T* array) {
      if (array)
        alignalloc::free(array);
    }
    
    //! zero a left-aligned memory block
    /**
       the block is aligned, but the length of it may not be a
       multiple of alignment. This zero routine does not touch
       anything beyond array+size.
    */
    template<class T> static inline void DINLINE zerola(T* array, size_t size) {
      // **** MMX/SSE memory zero routine
      //bzero(array,size*sizeof(T));
      mymemzerola(array,size*sizeof(T));
    }

  };//class alignalloc


/**
   template for allocating many single elements or small arrays of the
   same type with freeing them all together.

   Set bigblklen to define in how big blocks to allocate memory from
   the system. 
   
   If you allocate an array and the current big block has not enough
   space to accommodate the size, then the remaining space will be
   lost and a new big block will be allocated.

   You must not allocate an array of size more than bigblklen. In that
   case no warning occurs but you get a memory area smaller than you
   expect.

   All the allocated memory will be freed when the object is
   destructed or when you explicitly call freeall().
*/
template<class T, int bigblklen> class blockalloc {
public:
  blockalloc() {
    cbigblk=NULL;
    cfree=0;
  }

  //! allocate several consecutive elements (i.e. an array)
  T* DINLINE allocate(int ns) {
    if ((int)cfree<ns) {
      cfree=bigblklen;
      cbigblk=new T[bigblklen];
      //alignalloc::alloc(cbigblk,bigblklen);
      bigblocks.push_back(cbigblk);
    }
    T* ret=cbigblk;
    cfree-=ns;
    cbigblk+=ns;
    return ret;
  }

  //! allocate a single element
  T* DINLINE allocate() {
    if (!cfree) {
      cfree=bigblklen;
      cbigblk=new T[bigblklen];
      //alignalloc::alloc(cbigblk,bigblklen);

      bigblocks.push_back(cbigblk);
    }
    T* ret=cbigblk;
    --cfree;
    ++cbigblk;
    return ret;
  }

  //! free all allocated blocks
  void freeall() {
    for (TYPENAME bigblocks_t::iterator i=bigblocks.begin();
	 i!=bigblocks.end(); ++i) {
       delete [] *i;
       //alignalloc::free(*i);

    }
    //be prepared for another call of freeall() or allocate()
    bigblocks.clear();
    cbigblk=NULL;
    cfree=0;
  }

  ~blockalloc() {
    freeall();
  }

private:
  typedef std::list<T*> bigblocks_t; 
  //! the big blocks
  bigblocks_t bigblocks;
  
  //! the current partial big block
  T* cbigblk;
  
  //! the remaining length of the current partial big block
  size_t cfree;
};

/**
   template for allocating many single elements of the same type with
   freeing them all together, or freeing them one-by-one for reuse.

   Note that to maintain the freelist, the individual elements will be
   made at least the size of a pointer. It is thus not efficient to
   use this class with a data type T of size less than sizeof(void*).

   Note that upon free (for reusal) the destructor of T will _not_ be
   called and the first sizeof(void*) bytes will be silently
   overwritten by the freelist pointer. If this is undesired, please
   use some other container. (Or change the union to a
   struct/class/inheritance/etc...).

   Set bigblklen to define in how big blocks to allocate memory from
   the system. 
   
   All the allocated memory will be freed when the object is
   destructed or when you explicitly call freeall().
*/
/*template<class T, int bigblklen> class singlealloc {
public:
  singlealloc() :freelist(NULL){
  }

  //! allocate a single element
  T* DINLINE allocate() {
    if (freelist) {
      T* ret = (T*)freelist;
      freelist=freelist->next;
      return ret;
    }
    return (T*) allocator.allocate();
  }

  //! free all allocated blocks
  void freeall() {
    freelist=NULL;
    allocator.freeall();
  }

  //! free a single element (put it on the freelist)
  void free(T* data) {
    newdata_t *d=static_cast<newdata_t*>(data);
    d->next=freelist;
    freelist=d;
  }

  ~singlealloc() {
    freeall();
  }

private:
  //! elementary allocation type. We need to store either a T or a pointer
  typedef union newdata_t {
    T data;
    newdata_t *next;
  } newdata_t;

  //! the actual allocation is handled by a blockalloc
  typedef blockalloc<newdata_t, bigblklen> realalloc_t;
  
  //! the acutal allocator
  realalloc_t allocator;
  
  //! list of free elements
  T* freelist;
}; //class singlealloc
*/

template<class WRP, int bigblklen> class singlewrpalloc {
public:
  typedef typename WRP::data_t data_t;
  singlewrpalloc() :freelist(NULL){
  }

  //! allocate a single element
  data_t* DINLINE allocate() {
    if (freelist) {
      WRP* ret = freelist;
      freelist=freelist->next;
      return ret->getData();
    }
    return allocator.allocate()->getData();
  }

  //! free all allocated blocks
  void freeall() {
    freelist=NULL;
    allocator.freeall();
  }

  //! free a single element (put it on the freelist)
  void free(data_t* data) {
    WRP *d=(WRP*) data;
    d->next=freelist;
    freelist=d;
  }

  ~singlewrpalloc() {
    freeall();
  }

private:
  //! elementary allocation type. We need to store either a T or a pointer
  typedef WRP newdata_t;

  //! the actual allocation is handled by a blockalloc
  typedef blockalloc<WRP, bigblklen> realalloc_t;
  
  //! the acutal allocator
  realalloc_t allocator;
  
  //! list of free elements
  WRP* freelist;
}; //class singlealloc


  template<class T> class singleunionwrapper {
  public:
    typedef T data_t;
    union {
      T data;
      singleunionwrapper *next;
    };
    T* DINLINE getData() {
      return (T*) this;
    }
    singleunionwrapper *& DINLINE getNext() {
      return next;
    }
  }; //singleunionwrapper

  template<class T> class singlestructwrapper : T {
  public:
    typedef T data_t;

    singlestructwrapper *next;

    T* DINLINE getData() {
      return this;
    }
    singlestructwrapper *& DINLINE getNext() {
      return next;
    }
  }; //singlestructwrapper

  template<class T, int bigblklen> class singlesalloc: public singlewrpalloc<singlestructwrapper<T>, bigblklen> {};

  template<class T, int bigblklen> class singleualloc: public singlewrpalloc<singleunionwrapper<T>, bigblklen> {};


/** 
    memory array allocator class

    * it can allocate arrays of size BLOCKLEN(n)
    * freed arrays are kept on a freelist
    * all memory is returned to the system upon destruction

*/
template<class T,int base, int bigblklen> class arrayalloc{
public:
  static  size_t DINLINE BLOCKLEN(int size)  {
    return (base << size);
  }

  //! allocate a BLOCKLEN(size) sized block
  T* DINLINE allocate(unsigned int size) {
    if (freelists.size()<=size) {
      freelists.resize(size+1);
    }
    T* ret;
    if (ret=freelists[size]) {
      //we have a cached block
      T** d=(T**)ret;
      freelists[size]=*d;
      return ret;
    } else {
      //allocate a new block      
      size_t ns=BLOCKLEN(size);
      return thealloc.allocate(ns);
    }
  }//allocate

  //! free a BLOCKLEN(size) sized block
  void DINLINE free(T* data,unsigned int size) {
    if (freelists.size()<=size) {
      freelists.resize(size+1);
    }
    T** d=(T**)data;
    *d=freelists[size];
    freelists[size]=data;
  }//free
  
  
  arrayalloc() :thealloc(){
  }

private:

  //! the list of freed small blocks
  std::vector<T*> freelists;

  //! the actual allocator
  blockalloc<T,bigblklen> thealloc;

};




  /**
     class to allocate a decreasing-sized stack of memory blocks


  */
  template <class BLOCK> class debugblockstack {
  public:
    void DINLINE pushAndGetBlock(size_t length, typename BLOCK::dataget_t &d) {
      blocks.push(length);
      getCurrBlock().getData(d);
    }

    void DINLINE freeBlock() {
      if (blocks.empty()) {
        log_err(0,"freeBlock when stack is empty!");
      }
      blocks.pop();
    }

    BLOCK& getCurrBlock() {
      return blocks.top();
    }

    ~debugblockstack() {
      while (!blocks.empty()) {
        log_dbg(0,"Freeing block");
        blocks.top().free();
        blocks.pop();
      }
    }

  private:
    class smemblock_t : public BLOCK {
    public:
      size_t length;
      smemblock_t(size_t _length): BLOCK(_length),length(_length){}
    };

    //! container type used to hold the memory block list
    typedef std::stack<smemblock_t> blocklist_t;
    
    //! the actual stack of memory blocks
    blocklist_t blocks;

  }; //class debugblockstack

  template <class BLOCK> class blockstack {
  public:
    void DINLINE ensureCurrBlock(size_t length) {
      if ((currblock==blocks.end())||(length>currblock->length)) {
        //allocate block
        length+=length>>1;//allocate somewhat more memory
        smemblock_t newblock(length);
        blocks.insert(currblock,newblock);
        --currblock;
      }
    }

    void DINLINE stepBlock() {
      ++currblock;
    }

    void DINLINE pushAndGetBlock(size_t length, typename BLOCK::dataget_t &d) {
      ensureCurrBlock(length);
      getCurrBlock().getData(d);
      stepBlock();
    }

    void DINLINE freeBlock() {
      --currblock;
    }

    BLOCK& getCurrBlock() {
      return *currblock;
    }

    blockstack() {
      currblock=blocks.end();
    }

    ~blockstack() {
      for (typename blocklist_t::iterator it=blocks.begin();it!=blocks.end();++it) {
        it->free();
      }
    }

  private:
    class smemblock_t : public BLOCK {
    public:
      size_t length;
      smemblock_t(size_t _length): BLOCK(_length),length(_length){}
    };

    //! container type used to hold the memory block list
    typedef std::list<smemblock_t> blocklist_t;
    
    //! the actual stack of memory blocks
    blocklist_t blocks;
    
    //! the current block
    TYPENAME blocklist_t::iterator currblock;

  }; //class blockstack


  /**
     Simple block type for class blockstack
     It allocates a single length sized block of type T.

     if prezero is set to true, then memory blocks will be zeroed when
     allocated. In this case if you take care to return only zeroed
     blocks, then you need not zero each time you request a block from
     the stack.

     If you set it to false, then each time you get a block from the
     stack, you must zero it (or assume it is uninitialized). In this
     case you can return blocks in arbitrary state.

  */

  template<class T, bool PREZERO> class stacksingleblock {
  public:
    stacksingleblock(size_t length) {
      if (PREZERO) {
        alignalloc::allocz(data,length);
      } else {
        alignalloc::alloc(data,length);
      }
    }

    typedef T* dataget_t;

    void getData(dataget_t & r) {
      r=data;
    }
    
    void free() {
      alignalloc::freeifnonnull(data);
      data=NULL;
    }

  private:
    T* data;
  };//class stacksingleblock

  /**
     Multiple block type for class blockstack

     It allocates a single length sized block of type T, plus calls
     the given PARENT. 

     The 'get' operation requires a pair of a pointer-ref and whatever
     the parent requires. You can set the parent to a singleblock to
     get a pair of blocks each time. You can set the parent to a
     stackmultiblock to get more than two blocks.

     You can also join blocks of different base type together. 

     Keep in mind that the joined blocks will be always of the same size!

  */

  template<class T, bool PREZERO, class PARENT> class stackmultiblock : PARENT {
  public:
    stackmultiblock(size_t length) : PARENT(length) {
      if (PREZERO) {
        alignalloc::allocz(data,length);
      } else {
        alignalloc::alloc(data,length);
      }
    }

    typedef std::pair<T*, typename PARENT::dataget_t> dataget_t;
    void DINLINE getData(dataget_t &r) {
      PARENT::getData(r.second);
      r.first=data;
    }

    void free() {
      PARENT::free();
      alignalloc::freeifnonnull(data);
      data=NULL;
    }

  private:
    T* data;
  };//class stackmultiblock


} //namespace bracz


#endif //ALLOCATORS_HPP_050402
