/*!
  \file   log.h
  \brief  logging of status, warnings and error messages

  Provides routines for displaying and/or logging status, giving warnings.

  Set DEBUG_LEVEL to your needs with compiler option or rewrite debug
  information processing. There is a slot to provide debug level
  selection based on the message source (domain), which is currently
  __FILE__. This is yet unused (i.e. constant), you may set the debug
  level based on the source file.
  
  \author RACZ Balazs
  \date 2003-01-19 */

 
#ifndef __LOG_H__
#define __LOG_H__

//common interface

//! append to log, level ERR+\p level, \p format and \p args... like in printf()
#define log_err( level,  ... )  LOG( __FILE__ , LEVEL_ERR+level, "E", __VA_ARGS__ )
// csere gany nevutkozes miatt
//! append to log, level WARN+\p level, \p format and \p args... like in printf()
#define log_warn( level,  ... )  LOG( __FILE__ , LEVEL_WARN+level, "W", __VA_ARGS__)
//! append to log, level PERF+\p level, \p format and \p args... like in printf()
#define log_perf( level, ...)  LOG( __FILE__ , LEVEL_PERF+level, "p", __VA_ARGS__ )
//! append to log, level STATUS+\p level, \p format and \p args... like in printf()
#define log_status( level, ... )  LOG( __FILE__ , LEVEL_STATUS+level, "s",  __VA_ARGS__ )
//! append to log, level INFO+\p level, \p format and \p args... like in printf()
#define log_info( level, ... )  LOG( __FILE__ , LEVEL_INFO+level, "i",  __VA_ARGS__ )
//! append to log, level PERF+\p level, \p format and \p args... like in printf()
#define log_info_det( level, ...)  LOG( __FILE__ , LEVEL_INFO_DET+level, "p", __VA_ARGS__ )
//! append to log, level DEBUG+\p level, \p format and \p args... like in printf()
#define log_dbg( level, ... )  LOG( __FILE__ , LEVEL_DBG+level, "d", __VA_ARGS__ )

//! Base debug level of errors
#define LEVEL_ERR	0
//! Base debug level of warnings
#define LEVEL_WARN	10
//! Base level of performance (statistics) messages
#define LEVEL_PERF	20
//! Base debug level of status messages
#define LEVEL_STATUS	30
//! Base debug level of informative messages
#define LEVEL_INFO	40
//! Second (datailed) level of info (statistics) messages
#define LEVEL_INFO_DET	45
//! Base debug level of debugging messages
#define LEVEL_DBG	50


//implementation

#include <stdio.h>

//! \brief append to log (generic)
/** 
    \param domain is a string literal, specifies the source of the message

    \param level is a number, the greater, the less important the log
    entry. loglevel specification will filter levels to less than the
    specified number.
    
    \param levelchr is a string literal denoting the level.
    \param format is a printf() like format string. end-of-line will be appended to it.
    \param arg are the arguments to the format string.
*/
#define LOG( domain, level, levelchr, ... ) {if (level<=debug_level(domain)) { fprintf(stderr, domain "(" levelchr ",%d):",__LINE__); fprintf(stderr, __VA_ARGS__); fprintf(stderr,"\n");}}

//! \brief defines the debug level as a function of the domain.
/** currently constant
 */
#define debug_level( x )  DEBUG_LEVEL

#ifndef DEBUG_LEVEL
#warning You should define DEBUG_LEVEL or expect full debugging output everywhere
//! You should define DEBUG_LEVEL externally (makefile, perhaps).
#define DEBUG_LEVEL 255
#endif

#endif //__LOG_H__
