/*!
  \file   maxvector.hpp
  \brief  A back-insertion sequence container with fixed maximum size

  \author RACZ Balazs
  \date   2005-04-03
*/

#ifndef MAXVECTOR_HPP_050403
#define MAXVECTOR_HPP_050403


//#include "common/allocators.cpp"
#include <algorithm>
/**
  \brief  A back-insertion sequence container with fixed maximum size

  It never copies the contents and never checks the boundaries.

  The constructors and destructors are not properly called when the
  size of the vector changes. All elements are constructed and
  destructed tigether with the container.

  reserve(n) reallocates the array to accommodate at most n elements, and preserves min(size(),n) elements of the array.

*/
template<class T> class maxvector {
private:
  T* data;
  size_t length;
public:
  typedef T value_type;

  maxvector() {
    length=0;
    data=NULL;
  }

  maxvector(size_t maxsize) {
    data=new T[maxsize];
    length=0;
  }

  ~maxvector() {
    delete [] data;
  }

  T& DINLINE  operator[](size_t n) {
    return data[n];
  }

  const T& DINLINE  operator[](size_t n) const {
    return data[n];
  }

  typedef T* iterator;
  typedef const T* const_iterator;

  iterator DINLINE begin()  {
    return data;
  }

  const_iterator DINLINE begin() const {
    return data;
  }


  iterator DINLINE end()  {
    return data+length;
  }

  const_iterator DINLINE end() const  {
    return data+length;
  }

  size_t DINLINE size() const {
    return length;
  }

  /*void DINLINE push_back(T &d) {
    data[length++]=d;
    }*/
  
  /*  template<class U> void DINLINE push_back(const U &u) {
    data[length++]=T(u);
    }*/

  void DINLINE push_back(const T & d) {
    data[length++]=d;
  }

  void DINLINE pop_back() {
    length--;
  }

  bool DINLINE empty() const {
    return (!length);
  }

  T& DINLINE front() {
    return (data[0]);
  }
  
  T& DINLINE back() {
    return (data[length-1]);
  }

  const T& DINLINE back() const {
    return (data[length-1]);
  }


  void DINLINE reserve(size_t n) {
    T* newdata = new T[n];
    length=std::min(length,n);
    for(size_t i=0;i<length;i++) {
      newdata[i]=data[i];
    }
    delete [] data;
    data = newdata;
  }

  void DINLINE resize(size_t n, const T& d) {
    if (!data)
      reserve(n);
    for (size_t i=length;i<n;++i) {
      data[i]=d;
    }
    length=n;
  }

  void DINLINE resize(size_t n) {
    if (!data)
      reserve(n);
    length=n;
  }

  void DINLINE clear() {
    length=0;
  }

        iterator DINLINE erase(iterator first,
			       const iterator last)
      {
	 while( first!=last )
	    erase(first);
	 return first;
      }

  iterator DINLINE erase(iterator pos)
  {
    T* pos2 = pos;
    --length;
    while(pos2 != data + length)
      {
        *pos2 = *(pos2+1);
        ++pos2;
      }
    return pos;
  }

};

#endif //MAXVECTOR_HPP_050403
