/*
 * set_difference.hpp
 *
 * Time-stamp: <05/05/18 17:33:17 lars>
 *
 */

#ifndef __SET_DIFFERENCE__
#define __SET_DIFFERENCE__ t

#include "common/debug.hpp"

/**
 * Compute set difference of two vectors interpreted as ascending sorted lists of values,
 * represented by input iterators. 
 * i.e., compute x \setminus y.
 *
 * This is an alternative implementation of the STL method set-difference that hopefully
 * is maginally faster.
 *
 * Contrary to the STL implementation, this function 
 * <ul>
 * <li> returns the modified iterator through its result iterator argument (and not as return value).
 * <li> does not make copies of the xiter and yiter iterators, but modifies them.
 * </ul>
 * These deviations from STL may be changed later if they do not have any impact on performance.
 *
 * @param xiter input iterator of vector to subtract from.
 * @param xend  end of input iterator of a vector to subtract from.
 * @param yiter input iterator of vector to subtract.
 * @param yend  end of input iterator of vector to subtract.
 * @param resultIter output iterator of vector to fill setdifference into.
 */
template <class InputIterator1, class InputIterator2, class OutputIterator>
inline 
void set_difference_fast(InputIterator1& xiter, const InputIterator1& xend,
			 InputIterator2& yiter, const InputIterator2& yend,
			 OutputIterator& resultIter) {
  // 1. initialize loop
  int itemx = *xiter;
  int itemy = *yiter;

  // 2. loop over all elements
  while (true) {
    if (itemx < itemy) {
      *resultIter++ = itemx; 
      ++xiter; if (xiter == xend) return;
      itemx = *xiter;
    } else if (itemx > itemy) {
      ++yiter; if (yiter == yend) break;
      itemy = *yiter;
    } else { // if (itemx == itemy) {
      ++xiter; if (xiter == xend) return;
      ++yiter; if (yiter == yend) break;
      itemx = *xiter;
      itemy = *yiter;
    }
  }
  // 3. copy remaining items of x:
  for ( ; xiter != xend; ++xiter)
    *resultIter++ = *xiter;
}


/**
 * Compute set difference of two vectors interpreted as ascending sorted lists of values,
 * represented by input iterators. 
 * i.e., compute x \setminus y.
 * Exit operation if the result size does exceed a given upper bound (EARLY_FILTERING).
 *
 * @param xiter input iterator of vector to subtract from.
 * @param xend  end of input iterator of a vector to subtract from.
 * @param yiter input iterator of vector to subtract.
 * @param yend  end of input iterator of vector to subtract.
 * @param resultIter output iterator of vector to fill setdifference into.
 * @param upperSizeBound maximal size of result. REQ: 0 <= lowerSizeBound <= xend-xiter
 * @return <tt>true</tt> if the result size does not exceed upperSizeBound.
 */
template <class InputIterator1, class InputIterator2, class OutputIterator>
inline 
bool set_difference_upper_size_bound(InputIterator1& xiter, const InputIterator1& xend,
				     InputIterator2& yiter, const InputIterator2& yend,
				     OutputIterator& resultIter, const int upperSizeBound) {
  assert(upperSizeBound >= 0);
  assert(upperSizeBound <= xend-xiter);

  // 1. initialize loop
  OutputIterator resultEnd = resultIter + upperSizeBound;
  int itemx = *xiter;
  int itemy = *yiter;

  // 2. loop over all elements
  while (true) {
    if (itemx < itemy) {
      if (resultIter == resultEnd) return false; // EARLY_FILTERING
      *resultIter++ = itemx; 
      ++xiter; if (xiter == xend) return true; 
      itemx = *xiter;
    } else if (itemx > itemy) {
      ++yiter; if (yiter == yend) break;
      itemy = *yiter;
    } else { // if (itemx == itemy) {
      ++xiter; if (xiter == xend) return true; 
      ++yiter; if (yiter == yend) break;
      itemx = *xiter;
      itemy = *yiter;
    }
  }
  // 3. copy remaining items of x:
  if (resultIter + (xend - xiter) > resultEnd) 
    return false; // EARLY_FILTERING
  for ( ; xiter != xend; ++xiter) 
    *resultIter++ = *xiter;

  return true;
}

/**
 * Compute set difference of two vectors interpreted as ascending sorted lists of values,
 * represented by input iterators. 
 * i.e., compute x \setminus y.
 * Exit operation if the result does not have a given minimal number of elements.
 *
 * @param xiter input iterator of vector to subtract from.
 * @param xend  end of input iterator of a vector to subtract from.
 * @param yiter input iterator of vector to subtract.
 * @param yend  end of input iterator of vector to subtract.
 * @param resultIter output iterator of vector to fill setdifference into.
 * @param lowerSizeBound minimal size of result. REQ: 2 <= lowerSizeBound <= xend-xiter. (There are no improvements for lowerSizeBound = 0 or 1.)
 * @return <tt>true</tt> if the result size is at least lowerSizeBound.
 */
template <class InputIterator1, class InputIterator2, class OutputIterator>
inline 
bool set_difference_lower_size_bound(InputIterator1& xiter, InputIterator1& xend,
				     InputIterator2& yiter, const InputIterator2& yend,
				     OutputIterator& resultIter, const int lowerSizeBound) {
  assert(lowerSizeBound >= 2);
  assert(lowerSizeBound <= xend-xiter);

  // 1. initialize loop
  OutputIterator resultLowerBound = resultIter + lowerSizeBound - 1;
  xend -= lowerSizeBound - 1;
  int itemx = *xiter;
  int itemy = *yiter;

  // 2. loop over all elements
  while (true) {
    if (itemx < itemy) {
      if (resultIter < resultLowerBound)
	++xend;
      *resultIter++ = itemx; 
      ++xiter; if (xiter == xend) return resultIter > resultLowerBound;
      itemx = *xiter;
    } else if (itemx > itemy) {
      ++yiter; if (yiter == yend) break;
      itemy = *yiter;
    } else { // if (itemx == itemy) {
      ++xiter; if (xiter == xend) return resultIter > resultLowerBound;
      ++yiter; if (yiter == yend) break;
      itemx = *xiter;
      itemy = *yiter;
    }
  }
  // 3. copy remaining items of x:
  if (resultIter < resultLowerBound) 
    xend += resultLowerBound - resultIter;
  for ( ; xiter != xend; ++xiter) 
    *resultIter++ = *xiter;

  return true;
}


/**
 * Compute the size of the set difference of two vectors interpreted as ascending sorted lists of values,
 * represented by input iterators. 
 * i.e., compute |x \setminus y|.
 *
 * @param xiter input iterator of vector to subtract from.
 * @param xend  end of input iterator of a vector to subtract from.
 * @param yiter input iterator of vector to subtract.
 * @param yend  end of input iterator of vector to subtract.
 * @return size of the setdifference of x by y.
 */
template <class InputIterator1, class InputIterator2>
inline 
int size_of_set_difference(InputIterator1& xiter, const InputIterator1& xend,
			   InputIterator2& yiter, const InputIterator2& yend) {
  // 1. initialize loop
  int itemx = *xiter;
  int itemy = *yiter;
  int resultSize = 0;

  // 2. loop over all elements
  while (true) {
    if (itemx < itemy) {
      resultSize++;
      ++xiter; if (xiter == xend) return resultSize;
      itemx = *xiter;
    } else if (itemx > itemy) {
      ++yiter; if (yiter == yend) break;
      itemy = *yiter;
    } else { // if (itemx == itemy) {
      ++xiter; if (xiter == xend) return resultSize;
      ++yiter; if (yiter == yend) break;
      itemx = *xiter;
      itemy = *yiter;
    }
  }
  // 3. copy remaining items of x:
  resultSize += xend - xiter;

  return resultSize;
}


/**
 * Compute the size of the set difference of two vectors interpreted as ascending sorted lists of values,
 * represented by input iterators. 
 * i.e., compute |x \setminus y|.
 * Exit operation if the size of the result exceeds a given upper bound (EARLY_FILTERING).
 *
 * @param xiter input iterator of vector to subtract from.
 * @param xend  end of input iterator of a vector to subtract from.
 * @param yiter input iterator of vector to subtract.
 * @param yend  end of input iterator of vector to subtract.
 * @param upperSizeBound maximal size of result. REQ: upperSizeBound < x.size()
 * @return size of the setdifference of x by y, if it is at most upperSizeBound, or upperSizeBound+1 if it is larger.
 */
template <class InputIterator1, class InputIterator2>
inline 
int size_of_set_difference_upper_size_bound(InputIterator1& xiter, const InputIterator1& xend,
					    InputIterator2& yiter, const InputIterator2& yend, 
					    int upperSizeBound) {
  // 1. initialize loop
  int itemx = *xiter;
  int itemy = *yiter;
  int resultSize = 0;

  // 2. loop over all elements
  while (true) {
    if (itemx < itemy) {
      if (resultSize == upperSizeBound) return upperSizeBound+1; // EARLY_FILTERING
      resultSize++;
      ++xiter; if (xiter == xend) return resultSize;
      itemx = *xiter;
    } else if (itemx > itemy) {
      ++yiter; if (yiter == yend) break;
      itemy = *yiter;
    } else { // if (itemx == itemy) {
      ++xiter; if (xiter == xend) return resultSize;
      ++yiter; if (yiter == yend) break;
      itemx = *xiter;
      itemy = *yiter;
    }
  }
  // 3. copy remaining items of x:
  resultSize += xend - xiter;
  if (resultSize > upperSizeBound)
    resultSize = upperSizeBound+1;

  return resultSize;
}

/**
 * Compute the isze of the set difference of two vectors interpreted as ascending sorted lists of values,
 * represented by input iterators. 
 * i.e., compute |x \setminus y|.
 * Exit operation if the result does not have a given minimal number of elements.
 *
 * @param xiter input iterator of vector to subtract from.
 * @param xend  end of input iterator of a vector to subtract from.
 * @param yiter input iterator of vector to subtract.
 * @param yend  end of input iterator of vector to subtract.
 * @param resultIter output iterator of vector to fill setdifference into.
 * @param lowerSizeBound minimal size of result. REQ: 2 <= lowerSizeBound <= xend-xiter. (There are no improvements for lowerSizeBound = 0 or 1.)
 * @return result size if it is at least lowerSizeBound, 0 otherwise.
 */
template <class InputIterator1, class InputIterator2>
inline 
int size_of_set_difference_lower_size_bound(InputIterator1& xiter, InputIterator1& xend,
					    InputIterator2& yiter, const InputIterator2& yend,
					    const int lowerSizeBound) {
  assert(lowerSizeBound >= 2);
  assert(lowerSizeBound <= xend-xiter);

  // 1. initialize loop
  int resultSize = 0;
  const int resultLowerBound = lowerSizeBound - 1;
  xend -= lowerSizeBound - 1;
  int itemx = *xiter;
  int itemy = *yiter;

  // 2. loop over all elements
  while (true) {
    if (itemx < itemy) {
      if (resultSize < resultLowerBound)
	++xend;
      resultSize++;
      ++xiter; if (xiter == xend) return resultSize > lowerSizeBound ? resultSize : 0;
      itemx = *xiter;
    } else if (itemx > itemy) {
      ++yiter; if (yiter == yend) break;
      itemy = *yiter;
    } else { // if (itemx == itemy) {
      ++xiter; if (xiter == xend) return resultSize > lowerSizeBound ? resultSize : 0;
      ++yiter; if (yiter == yend) break;
      itemx = *xiter;
      itemy = *yiter;
    }
  }
  // 3. copy remaining items of x:
  if (resultSize < resultLowerBound) 
    xend += resultLowerBound - resultSize;
  resultSize += (xend - xiter);

  return resultSize;
}




/**
 * Compute set difference of two containers interpreted as ascending sorted lists of values,
 * i.e., compute x \setminus y.
 *
 * @param x container to subtract from. REQ: x.size() > 0 and sorted in ascending order.
 * @param y container to subtract. REQ: x.size() > 0 and sorted in ascending order.
 * @param result output container to fill setdifference into. REQ: result != x,y and result.capacity() >= x.size().
 */
template <class InputContainer1, class InputContainer2, class OutputContainer>
inline 
void set_difference(const InputContainer1& x,  const InputContainer2& y, OutputContainer& result) {
  assert(x.size() > 0); // 1. make sure inputs are not trivial.
  assert(y.size() > 0);

  assert(&result != &x); // 2. make sure output and inputs are not intermingled.
  assert(&result != &y);

  assert(result.capacity() >= x.size()); // 3. make sure output capacity is sufficient. 

  result.resize(x.size()); // STL containers do not have unbounded iterators -- therefore we have to resize the container in advance.

  typename InputContainer1::const_iterator xiter = x.begin();
  typename InputContainer2::const_iterator yiter = y.begin();
  typename OutputContainer::iterator resultIter = result.begin();
  set_difference_fast(xiter, x.end(), yiter, y.end(), resultIter);
  result.resize(resultIter - result.begin());
}


/**
 * Compute set difference of two containers interpreted as ascending sorted lists of values,
 * i.e., compute x \setminus y.
 * Exit operation if the result exceeds a given maximal size (EARLY_FILTERING).
 *
 * @param x container to subtract from. REQ: x.size() > 0 and sorted in ascending order.
 * @param y container to subtract. REQ: y.size() > 0 and sorted in ascending order.
 * @param result output container to fill setdifference into. REQ: result != x,y and result.capacity() >= min(x.size(), upperSizeBound)
 * @param upperSizeBound maximal size of result. REQ: upperSizeBound < x.size()
 * @return <tt>true</tt> if the result has at most maximalLength elements.
 */
template <class InputContainer1, class InputContainer2, class OutputContainer>
inline 
bool set_difference_upper_size_bound(const InputContainer1& x,  const InputContainer2& y, OutputContainer& result,
				     const int upperSizeBound) {
  assert(x.size() > 0); // 1. make sure inputs are not trivial.
  assert(y.size() > 0);
  assert(upperSizeBound >= 0); 
  assert(upperSizeBound < x.size()); // otherwise EARLY_FILTERING is turned on in vain.

  assert(&result != &x); // 2. make sure output and inputs are not intermingled.
  assert(&result != &y);

  assert(result.capacity() >= upperSizeBound || result.capacity() >= x.size()); // 3. make sure output capacity is sufficient. 

  result.resize(x.size() < upperSizeBound ? x.size() : upperSizeBound); // STL containers do not have unbounded iterators -- therefore we have to resize the container in advance.

  typename InputContainer1::const_iterator xiter = x.begin();
  typename InputContainer2::const_iterator yiter = y.begin();
  typename OutputContainer::iterator resultIter = result.begin();
  bool hasMaximalSize = set_difference_upper_size_bound(xiter, x.end(), yiter, y.end(), resultIter, upperSizeBound);
  result.resize(resultIter - result.begin());

  return hasMaximalSize;
}


/**
 * Compute set difference of two containers interpreted as ascending sorted lists of values,
 * i.e., compute x \setminus y.
 * Exit operation if the result size does not exceed a given lower bound.
 *
 * @param x container to subtract from. REQ: x.size() > 0 and sorted in ascending order.
 * @param y container to subtract. REQ: y.size() > 0 and sorted in ascending order.
 * @param result output container to fill setdifference into. REQ: result != x,y and result.capacity() >= min(x.size(), upperSizeBound)
 * @param lowerSizeBound minimum size of result. REQ: upperSizeBound < x.size()
 * @return <tt>true</tt> if the result size is at least a given lower bound.
 */
template <class InputContainer1, class InputContainer2, class OutputContainer>
inline 
bool set_difference_lower_size_bound(const InputContainer1& x,  const InputContainer2& y, OutputContainer& result,
				     const int lowerSizeBound) {
  assert(x.size() > 0); // 1. make sure inputs are not trivial.
  assert(y.size() > 0);
  assert(lowerSizeBound >= 1);       // otherwise EARLY_FILTERING is turned on in vain.
  assert(lowerSizeBound <= x.size());

  assert(&result != &x); // 2. make sure output and inputs are not intermingled.
  assert(&result != &y);

  assert(result.capacity() >= x.size()); // 3. make sure output capacity is sufficient. 

  result.resize(x.size()); // STL containers do not have unbounded iterators -- therefore we have to resize the container in advance.

  typename InputContainer1::const_iterator xiter = x.begin(), xend = x.end();
  typename InputContainer2::const_iterator yiter = y.begin();
  typename OutputContainer::iterator resultIter = result.begin();
  bool hasMinimumSize = set_difference_lower_size_bound(xiter, xend, yiter, y.end(), resultIter, lowerSizeBound);
  result.resize(resultIter - result.begin());

  return hasMinimumSize;
}

/**
 * Compute the size of the set difference of two containers interpreted as ascending 
 * sorted lists of values,
 * i.e., compute |x \setminus y|.
 *
 * @param x container to subtract from. REQ: x.size() > 0 and sorted in ascending order.
 * @param y container to subtract. REQ: y.size() > 0 and sorted in ascending order.
 * @return size of the setdifference of x by y.
 */
template <class InputContainer1, class InputContainer2>
inline 
int size_of_set_difference(const InputContainer1& x,  const InputContainer2& y) {
  assert(x.size() > 0); // 1. make sure inputs are not trivial.
  assert(y.size() > 0);

  typename InputContainer1::const_iterator xiter = x.begin();
  typename InputContainer2::const_iterator yiter = y.begin();
  return size_of_set_difference(xiter, x.end(), yiter, y.end());
}

/**
 * Compute the size of the set difference of two containers interpreted as ascending 
 * sorted lists of values,
 * i.e., compute |x \setminus y|.
 * Exit operation if the result exceeds a given maximal size (EARLY_FILTERING).
 *
 * @param x container to subtract from. REQ: x.size() > 0 and sorted in ascending order.
 * @param y container to subtract. REQ: y.size() > 0 and sorted in ascending order.
 * @param upperSizeBound maximal size of result. REQ: upperSizeBound < x.size()
 * @return size of the setdifference of x by y, if it is at most upperSizeBound, or upperSizeBound+1 if it is larger.
 */
template <class InputContainer1, class InputContainer2>
inline 
int size_of_set_difference_upper_size_bound(const InputContainer1& x,  const InputContainer2& y, const int upperSizeBound) {
  assert(x.size() > 0); // 1. make sure inputs are not trivial.
  assert(y.size() > 0);
  assert(upperSizeBound < x.size()); // otherwise EARLY_FILTERING is turned on in vain.

  typename InputContainer1::const_iterator xiter = x.begin();
  typename InputContainer2::const_iterator yiter = y.begin();
  return size_of_set_difference_upper_size_bound(xiter, x.end(), yiter, y.end(), upperSizeBound);
}

/**
 * Compute the size of the set difference of two containers interpreted as ascending 
 * sorted lists of values,
 * i.e., compute |x \setminus y|.
 * Exit operation if the result does not exceed  a given minimum size (EARLY_FILTERING).
 *
 * @param x container to subtract from. REQ: x.size() > 0 and sorted in ascending order.
 * @param y container to subtract. REQ: y.size() > 0 and sorted in ascending order.
 * @param lowerSizeBound minimum size of result. REQ: 2 <= lowerSizeBound <= x.size()
 * @return size of the setdifference of x by y, if it is at least lowerSizeBound, or 0 if it is smaller.
 */
template <class InputContainer1, class InputContainer2>
inline 
int size_of_set_difference_lower_size_bound(const InputContainer1& x,  const InputContainer2& y, const int lowerSizeBound) {
  assert(x.size() > 0); // 1. make sure inputs are not trivial.
  assert(y.size() > 0);
  assert(lowerSizeBound >= 2); // otherwise EARLY_FILTERING is turned on in vain.
  assert(lowerSizeBound <= x.size());

  typename InputContainer1::const_iterator xiter = x.begin(),
    xend = x.end();
  typename InputContainer2::const_iterator yiter = y.begin();
  return size_of_set_difference_lower_size_bound(xiter, xend, yiter, y.end(), lowerSizeBound);
}

#endif
