/**
 * SparseBitmatrix_setdifference.hpp
 *   -- compute setdifferences of all rows of a sparse bitmatrix with a given one of its rows.
 */

#ifndef __SPARSEBITMATRIX_SETDIFFERENCE_HPP_
#define __SPARSEBITMATRIX_SETDIFFERENCE_HPP_ t

#include "eclat/SparseBitvector_setdifference.hpp"
// #include "SparseBitmatrix.hpp"
//#include "eclat/SparseBitmatrix_singleblock.hpp"
#include "common/debug.hpp"


/**
 * Compute the difference of a given row minus every other row (with higher index).
 * I.e., compute bmIn_row \setminus bmIn_row' for all row' > row.
 * Keep only rows with at least one element but a given maximal number of elements;
 * mark row indices of rows identical to the reference row that would lead to empty rows.
 * 
 * @param bmIn original bit matrix. Each row'>=row has length >= ... [[minimalDefect]].
 * @param row reference row to compute rest of.
 * @param bmOut output bit matrix to fill setdifference into. 
 *                Each row must have capacity >= min(length(row), maximalLength) (with EARLY_FILTERING).
 *                                or capacity >= length(row)                     (without EARLY_FILTERING).
 * @param maximalLength maximal number of elements in results.
 * @param identicalRowsIterator output iterator for row indices of rows identical to the reference row 
 * @return number of rows in output bit matrix.
 */
template <bool EARLY_FILTERING, bool REMOVE_IDENTICAL_ROWS, class OutputIterator>
int setdifferenceOfSparseBitmatrix(const SparseBitmatrix& bmIn, const int row, SparseBitmatrix& bmOut, 
				   const int maximalLength, OutputIterator identicalRowsIterator) {
  assert(row >= 0 && row < bmIn.nRows());

  // 1. setup loop.
  bmOut.setNRows(bmIn.nRows() - row - 1);
  const GET_ROW(row1, bmIn, row); // legacy: const SparseBitvector row1(bmIn.getRow(row));
  int tor = 0;
  int maxRowLength = 0;

  // 2. iterate over all other rows:
  for (int fromr = row+1; fromr < bmIn.nRows(); fromr++) {
    // const SparseBitvector row2(bmIn.getRow(fromr));
    // SparseBitvector rowOut(bmOut.getRow(tor));
    const GET_ROW(row2, bmIn, fromr);
    GET_ROW(rowOut, bmOut, tor);

    const bool hasMaximalLength = EARLY_FILTERING && row1.length() > maximalLength
      ? setdifference<EARLY_FILTERING>(rowOut, row1, row2, maximalLength)
      : setdifference<false>(rowOut, row1, row2, maximalLength);

    // b. compute setdifference and retain if the defect is large enough.
    if (hasMaximalLength) {
      if (REMOVE_IDENTICAL_ROWS && rowOut.length() == 0)
	*identicalRowsIterator++ = bmIn.rowLabel(fromr); 
      else {
	if (rowOut.length() > maxRowLength)
	  maxRowLength = rowOut.length();
	// bmOut.setRowLength(tor, rowOut.length());
	bmOut.setRowLabel(tor, bmIn.rowLabel(fromr));
	tor++;
      }
    }
  }

  bmOut.setNRows(tor);
  bmOut.setMaxRowLength(maxRowLength);
  return tor;
}



/**
 * Compute the difference of every other row (with higher index) minus a given row.
 * I.e., compute bmIn_row' \setminus bmIn_row for all row' > row.
 * Keep only rows with at least one element and at most a given maximal length;
 * mark row indices of rows identical to the reference row that would lead to empty rows.
 * 
 * @param bmIn original bit matrix. Each row'>row has length <= lenght(row) + maximalLength.
 * @param row reference row to compute rest by.
 * @param bmOut output bit matrix to fill intersection into.
 *                The i-th row must have capacity(i) >= min(length(row+1+i), maximalLength)  if EARLY_FILTERING
 *                                    or capacity(i) >= length(row+1+i)                      if not EARLY_FILTERING
 * @param maximalLength maximal length of output rows.
 * @param identicalRowsIterator output iterator for row indices of rows identical to the reference row 
 * @return number of rows in output bit matrix.
 */
template <bool EARLY_FILTERING, bool REMOVE_IDENTICAL_ROWS, class OutputIterator>
int setdifferenceBySparseBitmatrix(const SparseBitmatrix& bmIn, const int row, SparseBitmatrix& bmOut, 
				   const int maximalLength, OutputIterator identicalRowsIterator) {
  assert(row >= 0 && row < bmIn.nRows());

  // 1. setup loop.
  bmOut.setNRows(bmIn.nRows() - row - 1);
  const GET_ROW(row1, bmIn, row); // legacy: const SparseBitvector row1(bmIn.getRow(row));
  int tor = 0;
  int maxRowLength = 0;

  // 2. iterate over all other rows:
  for (int fromr = row+1; fromr < bmIn.nRows(); fromr++) {
    const GET_ROW(row2, bmIn, fromr); // legacy: const SparseBitvector row2(bmIn.getRow(fromr));
    GET_ROW(rowOut, bmOut, tor);      // legacy: SparseBitvector rowOut(bmOut.getRow(tor));

    const bool hasMaximalLength = EARLY_FILTERING && row2.length() > maximalLength
      ? setdifference<EARLY_FILTERING>(rowOut, row2, row1, maximalLength)
      : setdifference<false>(rowOut, row2, row1, maximalLength);

    // b. compute setdifference and retain if the defect is large enough.
    if (hasMaximalLength) {
      if (REMOVE_IDENTICAL_ROWS && rowOut.length() == 0)
	*identicalRowsIterator++ = bmIn.rowLabel(fromr); 
      else {
	if (rowOut.length() > maxRowLength)
	  maxRowLength = rowOut.length();
	// bmOut.setRowLength(tor, rowOut.length());
	bmOut.setRowLabel(tor, bmIn.rowLabel(fromr));
	tor++;
      }
    }
  }

  bmOut.setNRows(tor);
  bmOut.setMaxRowLength(maxRowLength);
  return tor;
}


/**
 * Special case: only two rows. 
 * Compute |row1 \ row0| if <= maximalLength; return maximalLength + 1 otherwise.
 */
template <bool EARLY_FILTERING>
int cardinalitySetdifferenceBySparseBitmatrix_TwoRows(const SparseBitmatrix& bmIn, const int maximalLength) {
  const GET_ROW(row0, bmIn, 0); // legacy: const SparseBitvector row0(bmIn.getRow(0));
  const GET_ROW(row1, bmIn, 1); // legacy: const SparseBitvector row1(bmIn.getRow(1));
  return EARLY_FILTERING && row1.length() > maximalLength
    ? cardinalityOfSetdifference<EARLY_FILTERING>(row1, row0, maximalLength)
    : cardinalityOfSetdifference<false>(row1, row0, maximalLength);
}


#endif
