#ifndef DFDecoderWithEEManagement_HPP
#define DFDecoderWithEEManagement_HPP

/**
 * Writes patterns (itemsets plus support count) to disk.
 * <ul>
 * <li>Uses CacheDFDecoder for all i/o stuff (see there).
 * <li>Can manage a stack of equisupport extension items: <tt>write</tt> outputs
 *     the actual pattern as well as all its equisupport extensions
 *     as represented by the actual ee stack 
 *     (so be careful to push ee items *before* issueing the <tt>write</tt> method).  
 * <li>Keeps track of the state of the ee stack for each usual item 
 *     so there is no need for a method popEquisupportItem !
 * </ul>
 * Usage is as follows:
 * <ol>
 * <li>pushItem
 * <li>pushEEItem
 * <li>write
 * <li>popItem
 * </ol>
 * Please note that EE items have to be pushed *after* the respective item.
 * 
 * <p>It is possible to push EE items to the empty set (i.e. w.o. an initial pushItem).
 * These item *cannot* be poped again later on.
 *
 * @author Lars Schmidt-Thieme
 */

#include "CacheDFDecoder.hpp"

#include "common/debug.hpp"
#include <vector>
#include <iostream>
#include <iterator>
#include "io/output/OutputBase.hpp"

template <class DF = CacheDFDecoder< > >
class DFDecoderWithEEManagement : public DF
{
public:
  
      typedef typename DF::params_t params_t;
  
  DFDecoderWithEEManagement(const params_t* par) :
    DF(par)
  {
    ee_buffer_segment_length.push_back(0);
  }

      size_t nrOfEEItems() const
      {
	 return ee_buffer.size();
      }
      bool EEListEmpty() const
      {
	 return ee_buffer.empty();
      }
  /**
   * Push an item to the item stack.  
   *
   * note that functionality is duplicated in function
   * pushItemWithPrevSupport.
   */
  void pushItem(item_t item)
  {
#if DEBUG_LEVEL >= LEVEL_DBG
  std::cout << "pushItem " << item << std::endl; 
#endif
    DF::pushItem(item);
    ee_buffer_segment_length.push_back(0);
  }

  /**
   * Pop an item from the item stack.
   */
  void popItem()
  {
#if DEBUG_LEVEL >= LEVEL_DBG
    //assert(ee_buffer_segment_length.size()>0);
    std::cout << "popItem " << ee_buffer_segment_length.size() << " and ";
#endif
    // popEquisupportItems(ee_buffer_segment_length.back());
    popEquisupportItems(ee_buffer_segment_length.back());
    DF::popItem();
    ee_buffer_segment_length.pop_back();
  }

  /**
   * Pushes a given item on the stack of equisupport items.
   */
  void pushEquisupportItem(item_t item);

  /**
   * Pushes a given itemset on the stack of equisupport items.
   */
  void pushEquisupportItemSet(const std::vector<item_t>& itemset);


  /**
   * Writes actual itemset and all its equisupport extensions as represented by the  
   * actual stack of equisupport items. 
   */
  void write(counter_t support) {
    // as is -- w./o. any ee-extension (= empty ee set):
    DF::write(support);
    write1();
  }


  /**
   * Pushes an item on the item stack and writes actual itemset and
   * all its equisupport extensions as represented by the actual stack
   * of equisupport items. Support will be the same as the previous
   * write() or pushItemWithPrevSupport call set it to.
   */
  void pushItemWithPrevSupport(item_t item) {
    // as is -- w./o. any ee-extension (= empty ee set):
    DF::pushItemWithPrevSupport(item);
    ee_buffer_segment_length.push_back(0);
    write1();
  }

  /**
   * Pushes an item on the item stack and writes actual itemset and all its equisupport 
   * extensions as represented by the actual stack of equisupport items. 
   */
  void pushItemWithWrite(item_t item, counter_t support)
  {   
    pushItem(item);
    write(support);
  }

  /**
   * An interface for writing equisupport items:
   */ 
  class EquisupportItemOutputIterator {
  public:
    EquisupportItemOutputIterator(DFDecoderWithEEManagement& decoder) : decoder(decoder) {}
    EquisupportItemOutputIterator& operator= (const item_t item) { decoder.pushEquisupportItem(item); return *this; }
    EquisupportItemOutputIterator& operator* () { return *this; }
    EquisupportItemOutputIterator& operator++ () { return *this; }
    EquisupportItemOutputIterator& operator++ (int) { return *this; }
  private:
    DFDecoderWithEEManagement& decoder;
  };
  
private:
  /// The buffer for the equisupport extensions
  std::vector<item_t> ee_buffer;
  
      /** size of ee-buffer segment for each item, i.e., 
	  number of ee-items added specific before this item.
      */
  std::vector<size_t> ee_buffer_segment_length;


  /**
   * Writes the EE extensions with the prevSupport writer of the parent.
  */
  void write1();

  /**
   * Pops last item from the stack of equisupport items.
   */
  void popEquisupportItem();

  /**
   * Pops given number of last items from the stack of equisupport items.
   */
  void popEquisupportItems(int numEEItems);

};


template <class DF> inline void DFDecoderWithEEManagement<DF>::
pushEquisupportItem(item_t item)
{
#if DEBUG_LEVEL >= LEVEL_DBG
  std::cout << "pushEE " << item << std::endl; 
#endif
  ee_buffer_segment_length.back()++;
  ee_buffer.push_back(item);
}

template <class DF> inline void DFDecoderWithEEManagement<DF>::
pushEquisupportItemSet(const std::vector<item_t>& itemset)
{
#if DEBUG_LEVEL >= LEVEL_DBG
   std::cout<< "pushEEItemset "<<std::endl; 
   std::copy( itemset.begin(), itemset.end(), 
	      std::ostream_iterator<item_t>(std::cout, " "));
   std::cout<<std::endl; 
#endif
  ee_buffer_segment_length.back()+=itemset.size();
  ee_buffer.insert(ee_buffer.end(), itemset.begin(), itemset.end() );
}

template <class DF> inline void DFDecoderWithEEManagement<DF>::
popEquisupportItem()
{
  assert(!ee_buffer.empty());
#if DEBUG_LEVEL >= LEVEL_DBG
  std::cout << "popEE " << ee_buffer.back() << std::endl; 
#endif
  ee_buffer.pop_back();
  ee_buffer_segment_length.back()--;
}

template <class DF> inline void DFDecoderWithEEManagement<DF>::
popEquisupportItems(int numEEItems)
{
  assert(int(ee_buffer.size()) >= numEEItems);
#if DEBUG_LEVEL >= LEVEL_DBG
  std::cout << "popEEs ";
  if (numEEItems ==  0)
    std::cout << "[empty]";
  for(int i = 0; i < numEEItems; i++) 
    std::cout << ee_buffer[ee_buffer.size()-1-i] << " ";
  std::cout << " [" << numEEItems << "]" << std::endl; 
#endif
  ee_buffer.resize(ee_buffer.size() - numEEItems);
  ee_buffer_segment_length.back() -= numEEItems;
}



template <class DF> inline void DFDecoderWithEEManagement<DF>::
write1()
{
  /// The length of the ee_buffer
  std::vector<item_t>::size_type ee_buffer_length = ee_buffer.size();

  if (ee_buffer_length == 0)
    return;

  std::vector<bool> isMember(ee_buffer_length, false);



  // set only containing (numItems-1)
  isMember[ee_buffer_length-1] = true;
  DF::pushItemWithPrevSupport(ee_buffer[ee_buffer_length-1]);
  DF::popItem();

  int pos = ee_buffer_length-2;

  while (pos >= 0) {
    // a. subset containing ee_buffer[pos] but items with larger ee-index:
    isMember[pos] = true;
    DF::pushItemWithPrevSupport(ee_buffer[pos]);
    
    // b. add item (numItems-1):
    DF::pushItemWithPrevSupport(ee_buffer[ee_buffer_length-1]);

    // c. search next non-member (and remove all items on our way):
    pos = ee_buffer_length-2;
    while (pos >= 0 && isMember[pos]) {
      isMember[pos] = false;
      pos--;
    }
    // popEquisupportItems(ee_buffer_length-1 - pos);
    for (int i = ee_buffer_length-1 - pos;  i > 0; i--)
      DF::popItem();
  }
}

#endif
