#ifndef TransactionReader_HPP
#define TransactionReader_HPP

/**
  *@author Ferenc Bodon
  */

#include "common.hpp"  
#include "io/FDRepr.hpp"  
#include <iostream>  

/** This class represents the base class of transaction reader.

The transaction reader is responsible to read in transactions from a file.
*/

template <class FR = FDRepr>
class TransactionReader : public FR
{
   public:
      class params_t : public FR::params_t
      {
	 public:
	    /// the code of the largest item
	    item_t largest_item;

	    params_t() : FR::params_t()
	    {
	       largest_item = 0;
	    }
      };

      /**
	 \exeption ios_base::failure this exception is thrown when the input
	 file can not be read
      */
      TransactionReader( const params_t* par )	 
	 throw(std::ios_base::failure) : 
	 FR(par), largest_item(par->largest_item){}

      /** Reads in one transaction from the file and writes it to a class 
        whose type is given by the template.
       
        \param transaction this will be filled up wth the transaction 
        found in the next line of the file.
        \return \c 0 if the end of file is reached, otherwise \c the 
	multiplicity of the read transaction. In the case of end of
	file, the transaction will be empty.

        The template class must have a \code push_back \endcode  
        method that inserts an item to the transaction. Hence 
        <tt> stl::vector<item_t> </tt> or any 
        <tt> stl::Back Insertion Sequencce </tt> template class is applicable.
	An other restriction is that the template class must have a 
	\code clear \endcode  method that clears its content.
        
        \note
        If the type of the template is a 
        <tt>stl::Back Insertion Sequencce </tt> then the \a
        transaction will not be sorted and it will not be checked,
        if the \a transaction contains duplicates.<br>

	Also note that the \c transaction will always be overwritten. 
	For this the function \code clear() \endcode will be called.


       */
      template <class BIS> counter_t nextTransactionBIS( BIS& transaction )
      {
	 std::cerr<<"Error: using function nextTransactionBIS"<<std::endl;
         return 0;
      }

      /** Reads in one transaction from the file and writes it to a class 
        whose type is given by the template.
       
        \param transaction this will be filled up by the transaction 
        found in the next line of the file.
	\return \c 0 if the end of file is reached, otherwise \c the 
	multiplicity of the read transaction. In the case of end of
	file, the transaction will be empty.

        The template class must have a \code insert \endcode  
        method that inserts an item to the transaction. Hence 
        <tt> stl::set<item_t> </tt> or any <tt>stl::Unique 
        Associative Container</tt> template class is applicable.
	An other restriction is that the template class must have a 
	\code clear \endcode  method that clears its content.
        
        \note
        If the type of the template is a 
        <tt>stl::Unique Associative Container </tt> then the \a
        transaction is obviously ascendingly sorted, and no
        duplicates occur.<br>

	Also note that the \c transaction will always be overwritten. 
	For this the function \code clear() \endcode will be called.
       */
      template <class UAC> counter_t nextTransactionUAC( UAC& transaction )
      {
	 std::cerr<<"Error: using function nextTransactionUAC"<<std::endl;
         return 0;
      }


      /** It is possible to associate the largerst item in the transaction file
	  with the transaction reader.

	  This method is just for the sake of the convenience (to avoid many
	  parameter passing)
       */
      void setLargestItem(item_t largest_item)
      {this->largest_item = largest_item;}

      /** It returns the largest item
	  
      \note
      You get the value that you have last set by the setLargestItem function.
      If the setLargestItem function was never called, then the return value 
      is unspecified.
       */
      item_t getLargestItem() const
      {return largest_item;}

   protected:
      item_t largest_item;
};

#endif
