/** \mainpage FIM testing environment

Frequent Itemset Mining (FIM) is the most researched field of frequent
pattern mining. Over one hundred FIM algorithms were proposed - the
majority claiming to be the most efficient. To clarify this chaos and
the contradictions, two FIMI competitions were organized. It not only
resulted in some excellent implementations but the community also get
better understanding (limits, drawbacks, advantages) of the
algorithms.

The separate work of different research groups, however, cannot be so
efficient than tightly cooperating groups that interchange their
knowledge. The goal of this FIM environment is to support this tight
cooperation by providing an environment that includes base library
functions, like IO, coding, decoding, ... and some high level
desciption of the most important algorithms, such as apriori, eclat,
FP-growth together with some very efficient implementations of
them. The enviroment also contains many test classes to test the
solutions for specific problems.

Flexibility and plugability are achieved by a template-based solution,
that does not prevent the compiler form code optimization (like
virtual function-based solution), which results in codes that are
competitive with the best solutions.
 */


#include "common.hpp"
#include "common/log.h"
#include "io/input/transaction_reader/brBufferedTransactionReader.hpp"
#include "io/input/transaction_reader/SortedTransactionReader.hpp" 


#include "io/codec/decoder/df/SimpleDFDecoder.hpp"
//#include "io/codec/decoder/df/CacheDFDecoder.hpp"
#include "io/codec/decoder/df/DFDecoderWithEEManagement.hpp"
//#include "io/output/StatOutput.hpp"
#include "io/output/normal/SortOutput.hpp"
//#include "io/codec/decoder/df/DFDecoderWrapper.hpp"

#include "util/StreamParser.hpp"
#include "util/FrequentFilter.cpp"

#include "datastructures/maxvector.hpp"
#include "datastructures/trie/edgelist/OrderedEdgelist.hpp"
//#include "datastructures/trie/edgelist/OrderedEdgelistDynLookup.hpp"
#include "datastructures/trie/edgelist/OffsetIndexVector.hpp"

#include "apriori/bodon/Leaf.hpp"
#include "apriori/bodon/Trie.hpp"
#include "apriori/bodon/TrieNEE.hpp"
#include "io/input/transaction_reader/SortedTransactionReader.hpp"
#include "io/input/transaction_reader/OrderReverser.hpp"
#include "io/codec/coder/Coder.hpp"
#include "io/db_cache/BuildTreeDBCache.hpp"
#include "util/Frequent2Filter.cpp"
#include "util/Frequent2FilterOnline.cpp"
#include "apriori/OneByOneSupportCounter.hpp"
#include "apriori/bodon/dynamic_trie/trie_manipulators/SupportCounter.hpp"
//#include "apriori/bodon/dynamic_trie/trie_manipulators/SupportCounterMergeTries.hpp"
#include "apriori/AprioriSelector.hpp"
#include "fpgrowth/FPGrowthSelector.hpp"
#include <vector>
#include <iostream>
#include <string>


std::string file_format;

void init()
{
   file_format = "File format:";
   file_format += "\n\nThe transactionfile is a plan text file. Each row ";
   file_format += "represents a transaction. \n";
   file_format += "A transaction is a set of items seperated by a nonnumeric ";
   file_format += "character.\nIt can be for example a white space, comma, ";
   file_format += "colon, etc.\n";
   file_format += "Items are nonnegative integers.\n";
}
/// This procedure displays the usage of the program.
void usage()
{
   log_err(0,"Usage: fim algorithm transactionfile min_supp outcomefile [options]");
   log_err(0," algorithm\t    the name of the algorithm, i.e: apriori, eclat or fp-growth");
   log_err(0," transactionfile\t    file, that contains the tranasctions of items");
   log_err(0," outcomefile\t    file to write the outcome");
   log_err(0," min_supp\t    absolute support threshold");

   std::cerr << file_format;
   log_err(0,"\t\t\tHave a succesful mining ;-)\n\n");
}

/** This procedure processes the arguments.
   \return 
   <ul>
   <li>0, if no error is occured, </li>
   <li>1, in case of an IO error, </li>
   <li>2, in case of too few arguments,  </li>
   <li>3, if proper min_supp cannot be generated.  </li>
   <li>4, if proper maxsize cannot be generated.  </li>
   </ul>
 */

int process_arguments( int argc, char *argv[], counter_t& min_supp, 
		       bool &isrel, double &relminsupp, unsigned int& maxsize )
{
   if ( argc < 5 )
   {
     log_err(0,"There are 4 mandatory arguments: algorithm (apriori|fp-growth|eclat) inputfile min_supp (an integer, not percent) outputfile");
     return 2;
   }
   std::string mins=argv[3];
   if (mins[mins.size()-1]=='%') {
     mins.erase(mins.size()-1);
     isrel=true;
     relminsupp=atof(mins.c_str());
     relminsupp/=100;
     log_info(0,"Using relative minimum support of %lg",relminsupp);
     return 0;
   }
   isrel=false; relminsupp=0;
   int min_supp_i;
   try
   {
      convert(argv[3], min_supp_i);
      if ( min_supp_i <= 0  )
      {
	 log_err(0,"%s cannot be converted to a positive integer.",argv[3]);
	 return 3;
      }
   }
   catch(BadConversion e)
   {
      log_err(0,"min_supp conversion problem.");
      return 3;
   }
   min_supp = static_cast<counter_t>(min_supp_i);
   log_info(0,"min_supp is set to %d", min_supp);
   if(argc == 6)
   {
      int maxsize_i;
      try
      {
	 convert(argv[5], maxsize_i);
	 if ( maxsize_i <= 0  )
	 {
	    log_err(0,"%s cannot be converted to a positive integer.",argv[4]);
	    return 4;
	 }
      }
      catch(BadConversion e)
      {
	 log_err(0,"max_size conversion problem.");
	 return 4;
      }
      maxsize = static_cast<unsigned int>(maxsize_i);
      log_status(0,"maxsize is set to %d", maxsize);
   }
   else
      maxsize = largest_itemsetsize;
   return 0;
}

int main( int argc, char *argv[] )
{
   init();
   counter_t min_supp;
   unsigned int maxsize;
   bool relative;
   double relminsupp;
      
   {
      int return_val = 
	 process_arguments( argc, argv, min_supp, 
			    relative, relminsupp, maxsize );
      if(return_val)
	 return return_val;
   }

   char* algorithm = argv[1];
   char* input_file = argv[2];
   char* output_file = argv[4];

   try
   {
      // We assume that the transactions does not contain duplicates!!!
      typedef brBufferedTransactionReader< > T_R;
      // Otherwise uncomment this:
      // typedef SortedTransactionReader<brBufferedTransactionReader< >, true> T_R;

      T_R::params_t par_i;
      par_i.file_name = input_file;
      par_i.mode=FileReprBase::READ;
      par_i.file_buffer_size = 16 * 1024;
      T_R tr_reader(&par_i);
      std::vector< std::pair<counter_t, item_t> > freq_items_with_counters;
      counter_t nr_of_transactions;
      // The first step of each algorithms is determining the frequent items.
      FrequentFilter<T_R>
	 fr_filter(tr_reader);
      log_status(0,"Finding frequent items.");
      fr_filter.findFrequentItems( freq_items_with_counters,  
				   nr_of_transactions, min_supp);

      if(!freq_items_with_counters.empty())
      {
	 log_status(0,"Doing decoder.");
	 typedef DFDecoderWithEEManagement< > DF_D;
//	    typedef DFDecoderWithEEManagement< SimpleDFDecoder<SortOutput<> > > DF_D;
//	    typedef DFDecoderWithEEManagement< StatOutput<CacheDFDecoder< >, true> > DF_D;

	 DF_D::params_t par_d;
	 par_d.file_name = output_file;
	 par_d.mode=FileReprBase::WRITE;
//	    par_d.numfreq = freq_items_with_counters.size(); // If StatOutput is used!!!
	 DF_D df_decoder(&par_d);

	 if(strncmp(argv[1],"apriori",7) == 0)
	 {
	    log_status(0,"APRIORI is selected");
	    typedef Bodon::LeafWithoutConstructor LEAF_WC;	 
	    typedef Bodon::Leaf LEAF;	 
	    typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelist<std::vector<Edge> > > TRIE_OEL;
//	       typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelistDynLookup<std::vector<Edge>, 30 > > TRIE_OEL;
	    typedef Bodon::Trie< LEAF, Bodon::OffsetIndexVector< std::vector<void*> > > TRIE_OI;

	    typedef Bodon::TrieNEE<TRIE_OEL> TRIENEE_OEL;
	    typedef Bodon::TrieNEE<TRIE_OI> TRIENEE_OI;

	    const NEELevel NEE = NEE_Full;
	    log_status(0,"NEE_FULL is enabled");
/*	    if(strstr(argv[1],"mergetries"))
	    {
	       log_status(0,"Mergetrie option is on.");
	       typedef SortedTransactionReader< Coder<T_R, DF_D>, false> S_C_T_R;
	       const bool ENDONLY = false;
	       typedef bracz::BuildTreeDBCache< 
		  S_C_T_R, std::vector<item_t>, bracz::EndPatriciaBuildTree<ENDONLY>, ENDONLY, true > S_C;
	       typedef Bodon::dynamic_trie::SupportCounterMergeTries<TRIE_OEL, TRIE_OI, S_C> SUPP_C;
	       typedef Frequent2Filter<S_C> F2F;
	       AprioriSelector<TRIENEE_OEL, TRIENEE_OI, LEAF_WC, S_C, F2F, SUPP_C, T_R, DF_D, NEE>( 
		  min_supp, algorithm, input_file, nr_of_transactions, 
		  freq_items_with_counters, tr_reader, df_decoder, maxsize);

	    }
	    else*/ if(strstr(argv[1],"lowmem"))
	    {
	       log_status(0,"Low memory need option is on.");
	       typedef SortedTransactionReader<Coder<T_R, DF_D> >  S_C;
	       typedef Bodon::dynamic_trie::SupportCounter<TRIE_OEL, TRIE_OI> SUPP_C_BASE;
	       typedef OneByOneSupportCounter<TRIE_OEL, SUPP_C_BASE, S_C> SUPP_C;
	       typedef Frequent2FilterOnline<S_C> F2F;
	       AprioriSelector<TRIENEE_OEL, TRIENEE_OI, LEAF_WC, S_C, F2F, SUPP_C, T_R, DF_D, NEE>( 
		  min_supp, "apriori-intersect", input_file, nr_of_transactions, 
		  freq_items_with_counters, tr_reader, df_decoder, maxsize);
	    }
	    else 
	    {
	       typedef SortedTransactionReader< Coder<T_R, DF_D>, false, false > S_C_T_R;
	       const bool ENDONLY = true;
	       typedef OrderReverser< bracz::BuildTreeDBCache< 
	       S_C_T_R, std::vector<item_t>, bracz::EndPatriciaBuildTree<ENDONLY>, ENDONLY > >S_C;
	       typedef Bodon::dynamic_trie::SupportCounter<TRIE_OEL, TRIE_OI> SUPP_C_BASE;
	       typedef OneByOneSupportCounter<TRIE_OEL, SUPP_C_BASE, S_C> SUPP_C;
	       typedef Frequent2Filter<S_C> F2F;
	       AprioriSelector<TRIENEE_OEL, TRIENEE_OI, LEAF_WC, S_C, 
		  F2F, SUPP_C, T_R, DF_D, NEE>( 
		  min_supp, algorithm, input_file, nr_of_transactions, 
		  freq_items_with_counters, tr_reader, df_decoder, maxsize);
	    }
	 }
	 else if(strncmp(argv[1],"fp-growth",9) == 0)
	    FPGrowthSelector<T_R, DF_D>(
	       min_supp, relative, relminsupp, algorithm, input_file, 
	       nr_of_transactions, freq_items_with_counters, tr_reader, df_decoder);
	 else
	 {
	    usage();
	    log_err(0,"algorithm should be either apriori, fp-growth or eclat!");
	    return 1;
	 }
      }
   }
   catch (std::ios_base::failure e)
   {
      log_err(0,"Exiting the program due to IO exception");
      return 1;
   }
}


