#include "common.hpp"
#include "common/log.h"
#include "io/input/transaction_reader/brBufferedTransactionReader.hpp"
#include "io/input/transaction_reader/SortedTransactionReader.hpp" 


#include "io/codec/decoder/df/SimpleDFDecoder.hpp"
//#include "io/codec/decoder/df/CacheDFDecoder.hpp"
#include "io/codec/decoder/df/DFDecoderWithEEManagement.hpp"
//#include "io/output/StatOutput.hpp"
#include "io/output/normal/SortOutput.hpp"
//#include "io/codec/decoder/df/DFDecoderWrapper.hpp"

#include "util/StreamParser.hpp"
#include "util/FrequentFilter.cpp"

#include "datastructures/maxvector.hpp"
#include "datastructures/trie/edgelist/OrderedEdgelist.hpp"
//#include "datastructures/trie/edgelist/OrderedEdgelistDynLookup.hpp"
#include "datastructures/trie/edgelist/OffsetIndexVector.hpp"

#include "apriori/bodon/Leaf.hpp"
#include "apriori/bodon/Trie.hpp"
#include "apriori/bodon/TrieNEE.hpp"

#include "test/apriori/bodon/AprioriSelector.hpp"
#include <vector>
#include <iostream>
#include <string>


std::string file_format;

void init()
{
   file_format = "File format:";
   file_format += "\n\nThe transactionfile is a plan text file. Each row ";
   file_format += "represents a transaction. \n";
   file_format += "A transaction is a set of items seperated by a nonnumeric ";
   file_format += "character.\nIt can be for example a white space, comma, ";
   file_format += "colon, etc.\n";
   file_format += "Items are nonnegative integers.\n";
}
/// This procedure displays the usage of the program.
void usage()
{
   log_err(0,"Usage: candGendTester candidate_option[-nee_option] transactionfile min_supp outcomefile [maxsize]");
   log_err(0," candidate_option\t    the candidate generation technique, ");
   log_err(0,"i.e: simpleprune, intersectprune or noprune");
   log_err(0,"nee_option\t   the way equisupport items are handled");
   log_err(0,"i.e: nee (for prefix equisupport), neefull (for full equisupport pruning)");
   log_err(0," transactionfile\t    file, that contains the tranasctions of items");
   log_err(0," outcomefile\t    file to write the outcome");
   log_err(0," min_supp\t    absolute support threshold");
   log_err(0,"Example:\t candGendTester simpleprune ../data/kosarak.dat 1900 output.txt");
   log_err(0," \t candGendTester intersectprune-neefull ../data/kosarak.dat 1900 output.txt");

   std::cerr << file_format;
   log_err(0,"\t\t\tHave a succesful mining ;-)\n\n");
}

/** This procedure processes the arguments.
   \return 
   <ul>
   <li>0, if no error is occured, </li>
   <li>1, in case of an IO error, </li>
   <li>2, in case of too few arguments,  </li>
   <li>3, if proper min_supp cannot be generated.  </li>
   </ul>
 */

int process_arguments( int argc, char *argv[], counter_t& min_supp, 
		       bool &isrel, double &relminsupp, unsigned int& maxsize )
{
   if ( argc < 5 )
   {
     log_err(0,"There are 4 mandatory arguments.");
     return 2;
   }
   std::string mins=argv[3];
   if (mins[mins.size()-1]=='%') {
     mins.erase(mins.size()-1);
     isrel=true;
     relminsupp=atof(mins.c_str());
     relminsupp/=100;
     log_status(0,"Using relative minimum support of %lg",relminsupp);
     return 0;
   }
   isrel=false;
   int min_supp_i;
   try
   {
      convert(argv[3], min_supp_i);
      if ( min_supp_i <= 0  )
      {
	 log_err(0,"%s cannot be converted to a positive integer.",argv[3]);
	 return 3;
      }
   }
   catch(BadConversion e)
   {
      log_err(0,"min_supp conversion problem.");
      return 3;
   }
   min_supp = static_cast<counter_t>(min_supp_i);
   log_status(0,"min_supp is set to %d", min_supp);

   if(argc == 6)
   {
      int maxsize_i;
      try
      {
	 convert(argv[5], maxsize_i);
	 if ( maxsize_i <= 0  )
	 {
	    log_err(0,"%s cannot be converted to a positive integer.",argv[4]);
	    return 4;
	 }
      }
      catch(BadConversion e)
      {
	 log_err(0,"min_supp conversion problem.");
	 return 4;
      }
      maxsize = static_cast<unsigned int>(maxsize_i);
      log_status(0,"maxsize is set to %d", maxsize);
   }
   else
      maxsize = largest_itemsetsize;

   return 0;
}

int main( int argc, char *argv[] )
{
   init();
   counter_t min_supp;
   unsigned int maxsize;
   bool relative;
   double relminsupp;
      
   {
      int return_val = 
	 process_arguments( argc, argv, min_supp, 
			    relative, relminsupp, maxsize );
      if(return_val)
	 return return_val;
   }

   char* algorithm = argv[1];
   char* input_file = argv[2];
   char* output_file = argv[4];

   try
   {
      // We assume that the transactions does not contain duplicates!!!
      typedef brBufferedTransactionReader< > T_R;
      // Otherwise uncomment this:
      // typedef SortedTransactionReader<brBufferedTransactionReader< >, true> T_R;

      T_R::params_t par_i;
      par_i.file_name = input_file;
      par_i.mode=FileReprBase::READ;
      par_i.file_buffer_size = 16 * 1024;
      T_R tr_reader(&par_i);
      std::vector< std::pair<counter_t, item_t> > freq_items_with_counters;
      counter_t nr_of_transactions;
      // The first step of each algorithms is determining the frequent items.
      FrequentFilter<T_R>
	 fr_filter(tr_reader);
      log_status(0,"Finding frequent items.");
      fr_filter.findFrequentItems( freq_items_with_counters,  
				   nr_of_transactions, min_supp);

      if(!freq_items_with_counters.empty())
      {
	 log_status(0,"Doing decoder.");
	 typedef DFDecoderWithEEManagement< > DF_D;
//	    typedef DFDecoderWithEEManagement< SimpleDFDecoder<SortOutput<> > > DF_D;
//	    typedef DFDecoderWithEEManagement< StatOutput<CacheDFDecoder< >, true> > DF_D;

	 DF_D::params_t par_d;
	 par_d.file_name = output_file;
	 par_d.mode=FileReprBase::WRITE;
//	    par_d.numfreq = freq_items_with_counters.size(); // If StatOutput is used!!!
	 DF_D df_decoder(&par_d);

	 typedef Bodon::LeafWithoutConstructor LEAF_WC;	 
	 typedef Bodon::Leaf LEAF;	 
	 typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelist<std::vector<Edge> > > TRIE_OEL;
//	       typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelistDynLookup<std::vector<Edge>, 30 > > TRIE_OEL;
	 typedef Bodon::Trie< LEAF, Bodon::OffsetIndexVector< std::vector<void*> > > TRIE_OI;

	 if(strstr(argv[1],"-nee"))
	 {
//	       typedef Bodon::TrieNEE< Bodon::OrderedEdgelist<maxvector<Edge> > > TRIE;
	    typedef Bodon::TrieNEE<TRIE_OEL> TRIENEE_OEL;
	    typedef Bodon::TrieNEE<TRIE_OI> TRIENEE_OI;

	    if(strstr(argv[1],"-neefull"))
	    {
	       log_status(0,"NEE_FULL is enabled");
	       AprioriSelector<TRIENEE_OEL, TRIENEE_OI, LEAF_WC, T_R, DF_D, NEE_Full>( 
		  min_supp, algorithm, input_file, nr_of_transactions, 
		  freq_items_with_counters, tr_reader, df_decoder, maxsize);
	    }
	    else
	    {
	       log_status(0,"NEE_Level1 is enabled");
	       AprioriSelector<TRIENEE_OEL, TRIENEE_OI, LEAF_WC, T_R, DF_D, NEE_Level1>( 
		  min_supp, algorithm, input_file, nr_of_transactions, 
		  freq_items_with_counters, tr_reader, df_decoder, maxsize);
	    }
	 }
	 else
	 {
	    log_status(0,"NEE is disabled");
//	       typedef Bodon::Trie< Bodon::OrderedEdgelist<maxvector<Edge> > > TRIE;
	    AprioriSelector<TRIE_OEL, TRIE_OI, LEAF_WC, T_R, DF_D, NEE_Off>( 
	       min_supp, algorithm, input_file, nr_of_transactions, 
	       freq_items_with_counters, tr_reader, df_decoder, maxsize );

	 }
      }
   }
   catch (std::ios_base::failure e)
   {
      log_err(0,"Exiting the program due to IO exception");
      return 1;
   }
}


