#include "common.hpp"
#include "io/input/transaction_reader/brBufferedTransactionReader.hpp"
#include "io/input/transaction_reader/SortedTransactionReader.hpp"

#include "io/output/BufferedOutput.hpp"
#include "io/codec/decoder/df/CacheDFDecoder.hpp"

#include "util/StreamParser.hpp"
#include "util/FrequentFilter.cpp"

#include "datastructures/maxvector.hpp"

#include "test/apriori/bodon/ThresholdSelector.hpp"

#include <vector>
#include <iostream>
#include <string>


std::string file_format;

void init()
{
   file_format = "File format:";
   file_format += "\n\nThe transactionfile is a plan text file. Each row ";
   file_format += "represents a transaction. \n";
   file_format += "A transaction is a set of items seperated by a nonnumeric ";
   file_format += "character.\nIt can be for example a white space, comma, ";
   file_format += "colon, etc.\n";
   file_format += "Items are nonnegative integers.\n";
}
/// This procedure displays the usage of the program.
void usage()
{
   std::cerr << "\nUsage: findThreshold threshold vector_repr transactionfile min_supp ";
   std::cerr << "outcomefile [options]\n";
   std::cerr << "\n threshold\t  the threshold in command find, i.e: ";
   std::cerr << " the number of children above binary search is applied\n ";
   std::cerr << " vector_repr\t  the representation of the vector, i.e: ";
   std::cerr << "maxvector, stl::vector, \n";
   std::cerr << "\n transactionfile  file, that contains the transactions of items";
   std::cerr << "\n outcomefile\t  file to write the outcome";
   std::cerr << "\n min_supp\t  absolute support threshold";

   std::cerr << file_format;
   std::cerr << "\n\nHave a succesful mining ;-)"<<std::endl<<std::endl;
}

/** This procedure processes the arguments.
   \return 
   <ul>
   <li>0, if no error is occured, </li>
   <li>1, in case of an IO error, </li>
   <li>2, in case of too few arguments,  </li>
   <li>3, if proper min_supp cannot be generated.  </li>
   <li>4, if proper threshold cannot be generated.  </li>
   </ul>
 */

int process_arguments( int argc, char *argv[],
		       counter_t& min_supp, bool &isrel, double &relminsupp )
{
   if ( argc < 6 )
   {
      usage();
      log_err(0,"There are 5 mandatory arguments.");
      return 2;
   }
   std::string mins=argv[4];
   if (mins[mins.size()-1]=='%') {
      mins.erase(mins.size()-1);
      isrel=true;
      relminsupp=atof(mins.c_str());
      relminsupp/=100;
      log_info(0,"Using relative minimum support of %lg",relminsupp);
      return 0;
   }
   isrel=false;

   int min_supp_i;
   try
   {
      convert(argv[4], min_supp_i);
      if ( min_supp_i <= 0  )
      {
	 log_err(0,"%s cannot be converted to a positive integer.",argv[3]);
	 return 3;
      }
   }
   catch(BadConversion e)
   {
      log_err(0,"min_supp conversion problem.");
      return 3;
   }
   min_supp = static_cast<counter_t>(min_supp_i);
   log_info(0,"min_supp is set to %d", min_supp);
   return 0;
}

int main( int argc, char *argv[] )
{
   init();
   counter_t min_supp;
   bool relative;
   double relminsupp;
      
   {
      int return_val = 
	 process_arguments( argc, argv, min_supp, relative, relminsupp );
      if(return_val)
	 return return_val;
   }

   char* threshold = argv[1];
   char* input_file = argv[3];
   char* output_file = argv[5];

   try
   {
      // We assume that the transactions does not contain duplicates!!!
      typedef brBufferedTransactionReader< > T_R;
      // Otherwise uncmment this:
      // typedef SortedTransactionReader<brBufferedTransactionReader< >, true> T_R;

      T_R::params_t par_i;
      par_i.file_name = input_file;
      par_i.mode=FileReprBase::READ;
      par_i.file_buffer_size = 16 * 1024;
      T_R tr_reader(&par_i);
      std::vector< std::pair<counter_t, item_t> > freq_items_with_counters;
      counter_t nr_of_transactions;
      // The first step of each algorithms is determining the frequent items.
      FrequentFilter<T_R>
	 fr_filter(tr_reader);
      log_status(0,"Finding frequent items.");
      fr_filter.findFrequentItems( freq_items_with_counters,   
				   nr_of_transactions, min_supp);

      log_status(0,"Doing decoder.");
      typedef CacheDFDecoder< OutputBase<FDRepr> > DF_D;

      DF_D::params_t par_d;
      par_d.file_name = output_file;
      par_d.mode=FileReprBase::WRITE;
      DF_D df_decoder(&par_d);

      if(strcmp(argv[2],"maxvector")==0)
      {
	 log_info(0,"maxvector vector representation is selected");
	 ThresholdSelector<maxvector<Edge>, T_R, DF_D>( threshold, min_supp, input_file, nr_of_transactions, 
							freq_items_with_counters, tr_reader, df_decoder);
      }
      else if(strcmp(argv[2],"std::vector")==0)
      {
	 log_info(0,"std::vector vector representation is selected");
	 ThresholdSelector<std::vector<Edge>, T_R, DF_D>( threshold, min_supp, input_file, nr_of_transactions, 
							  freq_items_with_counters, tr_reader, df_decoder);
      }
      else
      {
	 usage();
	 log_err(0,"edge_repr should be either ordered_list, offsetindex!");
	 return 1;
      }
   }
   catch (std::ios_base::failure e)
   {
      log_err(0,"Exiting the program due to IO exception");
      return 1;
   }
}


