#include "common.hpp"
#include "common/log.h"
#include "common/allocators.hpp"
#include "io/input/transaction_reader/LBufferedTransactionReader.hpp"
#include "io/input/transaction_reader/SortedTransactionReader.hpp" 

#include "io/codec/coder/Coder.hpp"
#include "io/codec/decoder/df/SimpleDFDecoder.hpp"

#include "io/db_cache/bodon/RBTreeDBCache.hpp"

#include "util/StreamParser.hpp"
#include "util/FrequentFilter.cpp"
#include "util/Frequent2Filter.cpp"

#include "test/apriori/bodon/FattenableLeaf.hpp"
#include "apriori/bodon/Trie.hpp"
#include "datastructures/trie/edgelist/OrderedEdgelist.hpp"

#include "apriori/bodon/trie/trie_manipulators/FrequentItemInserter.hpp"
#include "apriori/bodon/trie/trie_manipulators/FrequentPairInserter.hpp"
#include "apriori/OneByOneSupportCounter.hpp"
#include "apriori/bodon/trie/trie_manipulators/support_counter/SupportCounterMerge.hpp"
#include "apriori/bodon/trie/trie_manipulators/SimplePruner.hpp"
#include "apriori/bodon/trie/trie_manipulators/CandidateGeneratorPrune.hpp"
#include "apriori/bodon/trie/trie_manipulators/InfreqRemover.hpp"

#include "apriori/Apriori.hpp"

#include <vector>
#include <iostream>
#include <string>


std::string file_format;

void init()
{
   file_format = "File format:";
   file_format += "\n\nThe transactionfile is a plan text file. Each row ";
   file_format += "represents a transaction. \n";
   file_format += "A transaction is a set of items seperated by a nonnumeric ";
   file_format += "character.\nIt can be for example a white space, comma, ";
   file_format += "colon, etc.\n";
   file_format += "Items are nonnegative integers.\n";
}
/// This procedure displays the usage of the program.
void usage()
{
   log_err(0,"Usage: apriori-simple transactionfile min_supp outcomefile [options]");
   log_err(0," transactionfile\t    file, that contains the tranasctions of items");
   log_err(0," outcomefile\t    file to write the outcome");
   log_err(0," min_supp\t    absolute support threshold");

   std::cerr << file_format;
   log_err(0,"\t\t\tHave a succesful mining ;-)\n\n");
}

/** This procedure processes the arguments.
   \return 
   <ul>
   <li>0, if no error is occured, </li>
   <li>1, in case of an IO error, </li>
   <li>2, in case of too few arguments,  </li>
   <li>3, if proper min_supp cannot be generated.  </li>
   </ul>
 */

int process_arguments( int argc, char *argv[], counter_t& min_supp, 
		       bool &isrel, double &relminsupp )
{
   if ( argc < 5 )
   {
     log_err(0,"There are 4 mandatory arguments.");
     return 2;
   }
   std::string mins=argv[2];
   if (mins[mins.size()-1]=='%') {
     mins.erase(mins.size()-1);
     isrel=true;
     relminsupp=atof(mins.c_str());
     relminsupp/=100;
     log_info(0,"Using relative minimum support of %lg",relminsupp);
     return 0;
   }
   isrel=false;
   int min_supp_i;
   try
   {
      convert(argv[2], min_supp_i);
      if ( min_supp_i <= 0  )
      {
	 log_err(0,"%s cannot be converted to a positive integer.",argv[3]);
	 return 3;
      }
   }
   catch(BadConversion e)
   {
      log_err(0,"min_supp conversion problem.");
      return 3;
   }
   min_supp = static_cast<counter_t>(min_supp_i);
   log_info(0,"min_supp is set to %d", min_supp);
   return 0;
}
template<class LEAF, class T_R, class DF_D> void 
helperFunction(
   T_R& tr_reader, DF_D& df_decoder, counter_t nr_of_transactions,
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters,
   char* input_file, counter_t min_supp)
{
   typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelist<std::vector<Edge> > > TRIE;
   TRIE main_trie;
   typedef SortedTransactionReader< Coder<T_R, DF_D> > S_C_T_R;
   typedef Bodon::RBTreeDBCache<S_C_T_R, std::vector<item_t> > S_C;
   typename S_C::params_t par_c;
   par_c.file_name = input_file;
   par_c.mode=FileReprBase::READ;
   par_c.largest_item = tr_reader.getLargestItem();
   par_c.decoder = &df_decoder;
   par_c.freq_items_with_counters = &freq_items_with_counters;
   par_c.codemode = ASC;
   log_status(0,"Doing sorted codec.");
   S_C sorted_coder(&par_c);
   
   std::vector< std::pair<counter_t, std::pair<item_t, item_t> > >
      freq_pairs_with_counters;
   Frequent2Filter<S_C> fr_2_filter(
      &sorted_coder );
//   Frequent2FilterOnline<S_C> fr_2_filter(
//      &sorted_coder );
   log_status(0,"Finding frequent pairs.")
      fr_2_filter.findFrequentPairs(freq_pairs_with_counters, min_supp);
   
   const NEELevel NEE =  NEE_Off;
   typedef NewWrapperAlloc<TRIE> TRIE_ALLOCATOR;
   typedef Bodon::FrequentItemInserter<DF_D, TRIE, NEE> FII;
   FII fii(main_trie, df_decoder);
   typedef Bodon::SupportCounterMerge<TRIE> SUPP_C_BASE;
   typedef OneByOneSupportCounter<TRIE, SUPP_C_BASE, S_C> SUPP_C;

   typedef Bodon::FrequentPairInserter<DF_D, TRIE, TRIE, TRIE_ALLOCATOR, NEE> FPI;
   typedef Bodon::trie::SimplePruner<DF_D, TRIE, NewWrapperAlloc<TRIE>, NEE> PRUNER;
   typedef Bodon::CandidateGeneratorPrune<PRUNER, DF_D, TRIE, TRIE_ALLOCATOR, NEE> CG;
   typedef Bodon::trie::InfreqRemover<DF_D, TRIE, TRIE_ALLOCATOR, NEE> IR;
   typedef Apriori<S_C, DF_D, TRIE, TRIE_ALLOCATOR, FII, FPI, CG, IR, SUPP_C> A;
   A apriori(main_trie, sorted_coder, df_decoder, fii);
   log_status(0,"Finding frequent itemsets.")
      apriori.findFrequentItemsets( 
	 nr_of_transactions, *par_c.freq_counters,
	 freq_pairs_with_counters, min_supp );
}


int main( int argc, char *argv[] )
{
   init();
   counter_t min_supp;
   bool relative;
   double relminsupp;
   
   {
      int return_val = 
	 process_arguments( argc, argv, min_supp, relative, relminsupp );
      if(return_val)
	 return return_val;
   }

   char* input_file = argv[1];
   char* output_file = argv[3];
   
   try
   {
      typedef LBufferedTransactionReader< > T_R;
      
      T_R::params_t par_i;
      par_i.file_name = input_file;
      par_i.mode=FileReprBase::READ;
      par_i.file_buffer_size = 16 * 1024;
      T_R tr_reader(&par_i);
      std::vector< std::pair<counter_t, item_t> > freq_items_with_counters;
      counter_t nr_of_transactions;
      // The first step of each algorithms is determining the frequent items.
      FrequentFilter<T_R>
	 fr_filter(tr_reader);
      log_status(0,"Finding frequent items.");
      fr_filter.findFrequentItems( freq_items_with_counters,  
				   nr_of_transactions, min_supp);
      
      if(!freq_items_with_counters.empty())
      {
	 log_status(0,"Doing decoder.");
	 typedef SimpleDFDecoder< > DF_D;

	 DF_D::params_t par_d;
	 par_d.file_name = output_file;
	 par_d.mode=FileReprBase::WRITE;
	 DF_D df_decoder(&par_d);

	 if(strcmp(argv[4], "0") == 0 )
	 {
	    log_status(0,"No vector is added to the nodes")
	    typedef Bodon::Leaf LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "1") ==0 )
	 {
	    log_status(0,"A vector of size 1 is added to each node")
	    typedef Bodon::FattenableLeaf<1> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "2") ==0 )
	 {
	    log_status(0,"A vector of size 2 is added to each node")
	    typedef Bodon::FattenableLeaf<2> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "4") ==0 )
	 {
	    log_status(0,"A vector of size 4 is added to each node")
	    typedef Bodon::FattenableLeaf<4> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "6") ==0 )
	 {
	    log_status(0,"A vector of size 6 is added to each node")
	    typedef Bodon::FattenableLeaf<6> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "8") ==0 )
	 {
	    log_status(0,"A vector of size 8 is added to each node")
	    typedef Bodon::FattenableLeaf<8> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "10") ==0 )
	 {
	    log_status(0,"A vector of size 10 is added to each node")
	    typedef Bodon::FattenableLeaf<10> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "11") ==0 )
	 {
	    log_status(0,"A vector of size 11 is added to each node")
	    typedef Bodon::FattenableLeaf<11> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "12") ==0 )
	 {
	    log_status(0,"A vector of size 12 is added to each node")
	    typedef Bodon::FattenableLeaf<12> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "13") ==0 )
	 {
	    log_status(0,"A vector of size 13 is added to each node")
	    typedef Bodon::FattenableLeaf<13> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "14") ==0 )
	 {
	    log_status(0,"A vector of size 14 is added to each node")
	    typedef Bodon::FattenableLeaf<14> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "15") ==0 )
	 {
	    log_status(0,"A vector of size 15 is added to each node")
	    typedef Bodon::FattenableLeaf<15> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "16") ==0 )
	 {
	    log_status(0,"A vector of size 16 is added to each node")
	    typedef Bodon::FattenableLeaf<16> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "17") ==0 )
	 {
	    log_status(0,"A vector of size 17 is added to each node")
	    typedef Bodon::FattenableLeaf<17> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "18") ==0 )
	 {
	    log_status(0,"A vector of size 18 is added to each node")
	    typedef Bodon::FattenableLeaf<18> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "19") ==0 )
	 {
	    log_status(0,"A vector of size 19 is added to each node")
	    typedef Bodon::FattenableLeaf<19> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "20") ==0 )
	 {
	    log_status(0,"A vector of size 20 is added to each node")
	    typedef Bodon::FattenableLeaf<20> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "21") ==0 )
	 {
	    log_status(0,"A vector of size 21 is added to each node")
	    typedef Bodon::FattenableLeaf<21> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "22") ==0 )
	 {
	    log_status(0,"A vector of size 22 is added to each node")
	    typedef Bodon::FattenableLeaf<22> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "23") ==0 )
	 {
	    log_status(0,"A vector of size 23 is added to each node")
	    typedef Bodon::FattenableLeaf<23> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "24") ==0 )
	 {
	    log_status(0,"A vector of size 24 is added to each node")
	    typedef Bodon::FattenableLeaf<24> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "26") ==0 )
	 {
	    log_status(0,"A vector of size 26 is added to each node")
	    typedef Bodon::FattenableLeaf<26> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "27") ==0 )
	 {
	    log_status(0,"A vector of size 27 is added to each node")
	    typedef Bodon::FattenableLeaf<27> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "28") ==0 )
	 {
	    log_status(0,"A vector of size 28 is added to each node")
	    typedef Bodon::FattenableLeaf<28> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "30") ==0 )
	 {
	    log_status(0,"A vector of size 30 is added to each node")
	    typedef Bodon::FattenableLeaf<30> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "31") ==0 )
	 {
	    log_status(0,"A vector of size 31 is added to each node")
	    typedef Bodon::FattenableLeaf<31> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "32") ==0 )
	 {
	    log_status(0,"A vector of size 32 is added to each node")
	    typedef Bodon::FattenableLeaf<32> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "34") ==0 )
	 {
	    log_status(0,"A vector of size 34 is added to each node")
	    typedef Bodon::FattenableLeaf<34> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "36") ==0 )
	 {
	    log_status(0,"A vector of size 36 is added to each node")
	    typedef Bodon::FattenableLeaf<36> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "40") ==0 )
	 {
	    log_status(0,"A vector of size 40 is added to each node")
	    typedef Bodon::FattenableLeaf<40> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
	 else if(strcmp(argv[4], "50") ==0 )
	 {
	    log_status(0,"A vector of size 50 is added to each node")
	    typedef Bodon::FattenableLeaf<50> LEAF;
	    helperFunction<LEAF, T_R, DF_D>(
	       tr_reader, df_decoder, nr_of_transactions,
	       freq_items_with_counters, input_file, min_supp );
	 }
      }
   }
   catch (std::ios_base::failure e)
   {
      log_err(0,"Exiting the program due to IO exception");
      return 1;
   }
}


