#ifndef InputreaderTester_HPP
#define InputreaderTester_HPP

/**
  *@author Ferenc Bodon
  */
#include "common.hpp"
#include "common/log.h"
#include "io/input/transaction_reader/TransactionReader.hpp"
#include "io/input/transaction_reader/LBufferedTransactionReader.hpp"
#include "io/input/transaction_reader/brBufferedTransactionReader.hpp"
#include "io/FileReprBase.hpp"
#include <vector>

#if DEBUG_LEVEL >= LEVEL_DBG
#include <iterator>
#endif

/** This class is for testing the input managers.

In this context testing means reading in all the transactions 
of a transactional database.  To select the best transaction reader
run testers/input-tester with different parameters, and measure the times, 
for example by using \c /usr/bin/time.

The read in transactions are neither stored nor used for anything. A temporal 
transaction is always overwritten with the new transaction. Also, this class does
not check if the transaction reader works properly. If you want to check,
compile the program with <tt> DEBUG_LEVEL >= LEVEL_DBG </tt>
(i.e. <tt>make input-reader DEBUG_LEVEL=255</tt>) and check if the output is the same
as the tranasction database. For example:
\code
./testers/input-tester stream simple ../data/T10I4D100K.dat 8000 1 | diff - ../data/T10I4D100K.dat
\endcode

Note that a transaction reader may change the order of item in a transaction, 
or in some databases the white space after the last items of all/some transactions
are ommited, and hence the output may differ from the transactional database.
*/

class InputreaderTester
{
   public:
      /// Prints out the useage of this class.
      void usageSimpleInput();


      /// The main function.
      void doTest( int argc, char *argv[]  );
   private:
      /// This function will select the reader used in the test.
      template <class F_R> void selectReader(
	 char* reader, char* file_name, unsigned int nr_of_scan,
	 std::streamsize lowlevel_buffer_size);

      /// This function will read in the transactions.
      template <class T_R> void readDatabase(
	 char* file_name, unsigned int nr_of_scan, 
	 std::streamsize lowlevel_buffer_size);
};
/**
   \param reader the name of the reader that has to be tested
   \param file_name the name of the file that contains 
   the transactions
   \param nr_of_scan the number the transaction datebase has to 
   be read through
   \param lowlevel_buffer_size the low level buffer size
 */
template <class F_R> void InputreaderTester::selectReader(
   char* reader, char* file_name, unsigned int nr_of_scan, 
   std::streamsize lowlevel_buffer_size)
{
   if( strcmp(reader, "lars") == 0 )
   {
      log_info(0,"Lars' buffered transaction reader is selected for test.");
      readDatabase< LBufferedTransactionReader< TransactionReader<F_R> > >(
	 file_name, nr_of_scan, lowlevel_buffer_size);
   }
   else if( strcmp(reader, "bracz") == 0 )
   {
      log_info(0,"Bracz' buffered transaction reader is selected for test.");
      readDatabase< brBufferedTransactionReader< TransactionReader<F_R> > >(
	 file_name, nr_of_scan, lowlevel_buffer_size);
   }
   else
      usageSimpleInput();   
}

/**
   \param file_name the name of the file that contains 
   the transactions
   \param nr_of_scan the number the transaction datebase has to 
   be read through
   \param lowlevel_buffer_size the low level buffer size
 */
template <class T_R> void InputreaderTester::readDatabase(
   char* file_name, unsigned int nr_of_scan, 
   std::streamsize lowlevel_buffer_size)
{
   typename T_R::params_t par;
   par.file_name = file_name;
   par.mode = FileReprBase::READ;
   par.file_buffer_size = lowlevel_buffer_size;

   try
   {
      T_R t_r(&par);
      std::vector<item_t> transaction;
      while(nr_of_scan)
      {
	 --nr_of_scan;
	 t_r.rewind();
	 log_info(0,"New scan started.");
	 while( t_r.nextTransactionBIS(transaction) )
	 {
#if DEBUG_LEVEL >= LEVEL_DBG
	    std::copy( transaction.begin(), transaction.end(), 
		       std::ostream_iterator<item_t>(std::cout, " "));
	    std::cout<<std::endl;
#endif
	 }
      }
   }
   catch (std::ios_base::failure e)
   {
      log_err(0,"Exiting the program due to IO exception");
      exit(1);
   }
}

#endif
