#ifndef FrequentFilter_CPP
#define FrequentFilter_CPP

/**
  *@author Ferenc Bodon
  */

#include "common.hpp"  
#include <vector>


/** This class determines the frequent items. It just reads
 in transactions one-by-one, increases the counter of the items, and 
 then selects the frequent items.
 \param IT_R the transaction reader
*/
template <class IT_R>
class FrequentFilter
{
   public:
      FrequentFilter<IT_R>(IT_R& it_r):it_r(it_r){}

      /**
	 if relative=true then relminsupp will be used to calculate the
	 absolute minimum support, which will be stored in min_supp.
	 relminsupp is a real number in [0,1]. (Thus not a percentage)
      */
      void findFrequentItems(
	 std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters,
	 counter_t& nr_of_transactions, counter_t& min_supp, 
	 bool relative=false, double relminsupp=0.0);
   private:
      IT_R& it_r;


};
/**
   \param freq_items_with_counters this vector will store the frequent items
   together with their counter. 
   \param nr_of_transactions the number of transactions in the database
   \param min_supp the support threshold
   \param relative true if the relative min_supp is used
   \param the relative min_supp value (if \c relative = \c true)
 */
template <class IT_R> void FrequentFilter<IT_R>::findFrequentItems(
   std::vector< std::pair<counter_t, item_t> >& freq_items_with_counters,
   counter_t& nr_of_transactions, counter_t& min_supp, 
   bool relative, double relminsupp)
{
   freq_items_with_counters.clear();
   nr_of_transactions = 0;
   it_r.rewind();

   std::vector<item_t> transaction;
   std::vector< counter_t > temp_counter_vector;

   /// Determining the support of the items
   std::vector<item_t>::iterator it_transaction;
   while( it_r.nextTransactionBIS( transaction ) )
   {      
      if( !transaction.empty() )
      {
	 nr_of_transactions++;
	 for( it_transaction = transaction.begin(); 
	      it_transaction != transaction.end(); ++it_transaction )
	 {
	    if( *it_transaction + 1  > temp_counter_vector.size() )
	       temp_counter_vector.resize( *it_transaction + 1, 0 );
	    ++temp_counter_vector[*it_transaction];
	 }
      }
   }
   it_r.setLargestItem( temp_counter_vector.size() - 1 );
      
   if(relative)
      min_supp = static_cast<unsigned int>(relminsupp * nr_of_transactions);
   /// Finding the frequent items
   for( std::vector< counter_t >::size_type index = 0; 
	index < temp_counter_vector.size(); ++index )
      if(temp_counter_vector[index] >= min_supp)
      {
	 std::pair<counter_t, item_t> temp_pair(temp_counter_vector[index],index);
	 freq_items_with_counters.push_back(temp_pair);
      }
}

#endif
